# Created: 11/16/2004
# This class is part of the 3DNAParser package

########
# The Bio.PDB code is reused and modified for the 3DNAParser package.
#
# Copyright (C) 2002, Thomas Hamelryck (thamelry@vub.ac.be)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.  
########

from string import join
from numpy import array, dot, conj
from math import sqrt

from Entity import ChildEntity, DisorderedEntityWrapper

__doc__ = "Atom class in Structure objects."

class Atom(ChildEntity):
    """
    Atom object.
    
    The Atom object stores atom name (both with and without spaces), 
    coordinates, B factor, occupancy, alternative location specifier
    and (optionally) anisotropic B factor and standard deviations of 
    B factor and positions.
    
    Arguments:
    o id - atom name, the id of the Atom object (spaces are normally stripped)
      type: string
    o counter - atom countering number
      type: integer
    o serial number - atom serial number
      type: integer
    o fullname - full atom name, including spaces, e.g. ' CA '.
      type: string
    o coord - atomic coordinates (x, y, z)
      type: array
    o bfactor - isotropic B factor
      type: float
    o occupancy - occupancy (0.0-1.0)
      type: float
    o altloc - alternative location specifier for disordered atoms
      type: string

    Other attributes (set in PDBParser):
    o anisou_array - anisotropic B factor
      type: array
    o siguij_array - standard deviation of anisotropic B factor
      type: array
    o sigatm_array - standard deviation of atom position
      type: array
    o disordered - flag of disordered atom, 1 or 0
      type: integer
    """
    def __init__(self, id, counter, serial_number, fullname, coord, bfactor, occupancy, altloc):
        self.level = "A"
        self.full_id = None
        self.counter = counter
        self.serial_number = serial_number
        self.fullname = fullname
        self.coord = coord
        self.bfactor = bfactor
        self.occupancy = occupancy
        self.altloc = altloc
        self.disordered = 0
        self.anisou_array = None
        self.siguij_array = None
        self.sigatm_array = None
        ChildEntity.__init__(self, id)

    # Special methods
    def __repr__(self):
        return "<Atom %s>" % self.get_id()

    def __sub__(self, other):
        """
        Calculate distance between two atoms.        
        Example:
            >>> distance = atom1 - atom2
        o other: the other atom
          type: L{Atom}
        o Return value: distance, float
        """
        vab = self.coord - other.coord
        return sqrt(dot(vab,vab))
            
    # Public methods
    # Set methods
    def set_serial_number(self, number):
        "Set the serial number."
        self.serial_number = number

    def set_bfactor(self, bfactor):
        "Set the bfactor."
        self.bfactor = bfactor

    def set_coord(self, coord):
        "Set the coordinates (type: array) of the atom."
        self.coord = coord

    def set_altloc(self, altloc):
        "Set the alternative location."
        self.altloc = altloc

    def set_occupancy(self, occupancy):
        "Set the occupancy."
        self.occupancy = occupancy

    def set_sigatm(self, sigatm_array):
        """
        Set standard deviation of atomic parameters.
        The standard deviation of atomic parameters consists
        of 3 positional, 1 B factor and 1 occupancy standard 
        deviation.
        o sigatm_array - standard deviations of atomic parameters.
          type: Numpy array (length 5)
        """
        self.sigatm_array = sigatom_array

    def set_siguij(self, siguij_array):
        """
        Set standard deviations of anisotropic temperature factors.
        o siguij_array - standard deviations of anisotropic temperature factors.
          type: Numpy array (length 6)
        """
        self.siguij_array = siguij_array

    def set_anisou(self, anisou_array):
        """
        Set anisotropic B factor.
        o anisou_array - anisotropic B factor.
          type: Numpy array (length 6)
        """
        self.anisou_array = anisou_array
        
    # Get methods
    def get_full_id(self):
        """
        Return the full id of the atom.
        The full id of an atom is a tuple 
        (structure id, model id, chain id, residue id, atom name, altloc).
        """
        return self.parent.get_full_id() + ((self.id, self.altloc),)
        
    def get_fullname(self):
        "Return the full atom name (with spaces)."
        return self.fullname

    def get_serial_number(self):
        "Return the serial number."
        return self.serial_number

    def get_coord(self):
        "Return the (x, y, z) coordinates of the atom (type: array)."
        return self.coord

    def get_bfactor(self):
        "Return the bfactor of the atom."
        return self.bfactor

    def get_occupancy(self):
        "Return the occupancy of the atom."
        return self.occupancy

    def get_altloc(self):
        "Return the alternative location of the atom."
        return self.altloc

    def get_sigatm(self):
        "Return standard deviation of atomic parameters."
        return self.sigatm_array

    def get_siguij(self):
        "Return standard deviations of anisotropic temperature factors."
        return self.siguij_array

    def get_anisou(self):
        "Return anisotropic B factor."
        return self.anisou_array

    def get_identity(self):
        """
        Identify the atom class if it belongs a nucleotide.
        o Return values:
          B - base atom
          S - sugar atom
          P - phosphate atom (including 'P', 'O1P', 'O2P')
        """            
        aname = upper(self.id)
        if re.search('^\w*P', aname):
            return "P"
        elif re.search('^\w[\*\']', aname):
            return "S"
        else:
            return "B"

    # Other public methods
    def destroy(self):
        "Destroy the atom."
        self.parent = None
        del self.id

    def flag_disorder(self):
        "Set the disordered flag to 1."
        self.disordered = 1
    
    def is_disordered(self):
        "Return 1 if the atom is disordered."
        return self.disordered


class DisorderedAtom(DisorderedEntityWrapper):
    """
    This class contains all Atom objects that represent the same disordered
    atom. One of these atoms is 'selected' and all method calls not caught
    by DisorderedAtom are forwarded to the selected Atom object. In that way, a
    DisorderedAtom behaves exactly like a normal Atom.

    ### old definition of disordered_select
    By default, the selected Atom object represents the Atom object with
    the highest occupancy, but a different Atom object can be selected by
    using the disordered_select(altloc) method.

    ### new definition
    The selected atom is the one with smallest value of altloc.
    """
    def __init__(self, id):
        """
        Arguments:
        o id - string, atom name
        """
        #self.last_occupancy = -1
        DisorderedEntityWrapper.__init__(self, id)

    # Special methods

    def __repr__(self):
        return "<Disordered Atom %s>" % self.get_id() 

    def disordered_add(self, atom):
        """
        Add a disordered atom.
        ### old definition of the selected atom:
        the atom with highest occupancy value.
        In order to be consistent with the altloc setting in 3DNA,
        change the selected atom to the atom with 'A' or '1' altloc code.
        """
        # Add atom to dict, use altloc as key   
        atom.flag_disorder()
        # set the residue parent of the added atom
        residue = self.get_parent()
        atom.set_parent(residue)
        altloc = atom.get_altloc()
        self[altloc] = atom
        # get sorted altloc list
        altloc_list = self.disordered_get_id_list()
        self.disordered_select(altloc_list[0])
