# Created: 11/16/2004
# This class is part of the 3DNAParser package

from string import join
from numpy import concatenate

from Entity import ChildEntity

__doc__ = "Class for base pair."

class Basepair(ChildEntity):
    """
    The base class for base pairs found in helices and unpaired regions.
    
    Arguments:
    o id - res_counter1, res_counter2
      type: tuple
    o counter - base-pair counter
      type: int
    o edge - interacting edge, e.g., W or S
      type: string
    o residue - reordered residue
      type: tuple of two residue objects
    o bp_para - base-pair parameters: Shear, Stretch, Stagger, Buckle,
                Propeller, Opening
      type: array
    o iso_para - isosteric parameters
      type: array
    o hbonds - hydrogen bonds, (atom1, atom2, hb_dist, hb_attrbt)
      type: tuple of tuple
    o hbond_num - the number of hbonds
      type: int
    o bp_order - the order of two nucleotides
      type: int
    o bp_label - bp label, original/reversed/special
      type: string
    o bp_type - class of base pair (multiple values), single/regular/bifurcated/symmetric
      type: tuple
    o orientation - base-pair orientation
      type: tuple
    o pat_id - basepair pattern id (refname1, refname2, bp_orie, hbonds)
      type: tuple
    o pat_name - pattern name, e.g., ga_4
      type: string
    o isopat_name - isosteric pattern name, e.g., 33
      type: int
    o multi_id - multi_id
      type: string
    o pdist - phosphate distance
      type: float
    o lw_orie - Leontis-Westhof base-pairing orientation
      type: string

    The order of residues in a base pair is standardized before initiation.
    Four steps 
    """
    def __init__(self, id, residue1, residue2, bp_para, iso_para,\
                 hbonds, hbond_num, bp_order, bp_label, bp_type, pat_id):
        self.counter = None
        self.edge1 = "-"
        self.edge2 = "-"
        self.residue1 = residue1
        self.residue2 = residue2
        self.bp_para = bp_para
        self.iso_para = iso_para
        self.hbonds = hbonds
        self.hbond_num = hbond_num
        self.bp_order = bp_order
        self.bp_label = bp_label
        self.bp_type = bp_type
        self.orientation = None
        self.pat_id = pat_id
        self.pat_name = None
        self.isopat_name = None
        self.multi_ids = []
        self.pdist = None
        self.lw_orie = "-"
        ChildEntity.__init__(self, id)

    # Special methods
    def __repr__(self):
        return "<Base pair: %s, %s>" % (self.residue1, self.residue2)

    # Public methods
    # Set methods
    def set_edges(self, edges):
        "Set two interacting edges of current base pair."
        self.edge1 = edges[0]
        self.edge2 = edges[1]

    def set_bp_para(self, bp_para):
        self.bp_para = bp_para
        
    def set_orientation(self, orientation):
        self.orientation = orientation

    def set_hbonds(self, hbonds):
        self.hbonds = hbonds

    def set_hbnum(self, hbond_num):
        self.hbond_num = hbond_num

    def set_pat_id(self, pat_id):
        self.pat_id = pat_id

    def set_pat_name(self, pat_name):
        self.pat_name = pat_name        

    def set_isopat_name(self, isopat_name):
        self.isopat_name = isopat_name
        
    def set_order(self, bp_order):
        self.bp_order = bp_order

    def set_label(self, bp_label):
        self.bp_label = bp_label
        
    def set_bp_type(self, bp_type):
        self.bp_type = bp_type

    def set_iso_para(self, iso_para):
        self.iso_para = iso_para

    def set_multi_id(self, multi_id):
        if multi_id not in self.multi_ids:
            self.multi_ids.append(multi_id)

    def set_pdist(self, pdist):
        self.pdist = pdist

    def set_lworie(self, lw_orie):
        self.lw_orie = lw_orie

    # Get methods
    def get_lworie(self):
        return self.lw_orie

    def get_coords(self):
        """
        Return coordinates of two residues in array form.
        """
        residue1 = self.residue1
        residue2 = self.residue2
        rcoord1 = residue1.get_coord_array()
        rcoord2 = residue2.get_coord_array()
        return concatenate((rcoord1, rcoord2))

    def get_edges(self):
        "Return the pair of edges."
        return self.edge1, self.edge2
    
    def get_respair(self):
        "Return the pair of residues in order."
        return self.residue1, self.residue2

    def get_bp_para(self):
        "Return the array of base-pair parameter."
        return self.bp_para

    def get_orientation(self):
        "Return the orientation list of the base pair."
        return self.orientation

    def get_hbonds(self):
        "Return the hydrogen bonding pattern of the base pair."
        return self.hbonds

    def get_hbnum(self):
        "Return the hydrogen-bond number."
        return self.hbond_num

    def get_pat_id(self):
        "Return the base-pairing pattern id."
        return self.pat_id

    def get_order(self):
        "Return the order of two residues in the base pair."
        return self.bp_order

    def get_label(self):
        "Return the class of the base pair."
        return self.bp_label
    
    def get_bp_type(self):
        return self.bp_type

    def get_iso_para(self):
        "Return the isosteric parameters of the base pair."
        return self.iso_para

    def get_isopat_id(self):
        "Return the isosteric pattern id (edge1, edge2, bp_orie)."
        return (self.edge1, self.edge2, self.lw_orie)

    def get_counter(self):
        "Return the base-pair contering number."
        return self.counter
    
    def get_pat_name(self):
        return self.pat_name

    def get_isopat_name(self):
        return self.isopat_name

    def get_pdist(self):
        return self.pdist

    def get_multi_ids(self):
        return self.multi_ids
