# Created: 3/16/2005
# This class is part of the 3DNAParser package.
# 
# The 3DNABuilder class was to initiate objects for 3DNAInforer.
#
# The StructureBuilder class was modified from the Bio.PDB.PDBParser class.
#
########
# The Bio.PDB code is reused and modified for the 3DNAParser package.
#
# Copyright (C) 2002, Thomas Hamelryck (thamelry@vub.ac.be)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.  
########

import sys
import re
from math import pi, sin, cos, atan2, degrees, floor
from string import upper, lower, join, strip, split
from numpy import array, zeros

from Structure import Structure
from Model import Model
from Chain import Chain
from Residue import Residue
from Atom import Atom, DisorderedAtom

from X3DNAInformer import X3DNAInformer
from Secondary import Secondary
from Helix import Helix
from Unpaired import Unpaired
from BasepairH import BasepairH
from Tertiary import Tertiary
from BasepairU import BasepairU
from StepWrapper import StepWrapper
from SHelix import SHelix
from Basestep import Basestep
from MultiWrapper import MultiWrapper
from Multiplet import Multiplet

from Isostere import Isostere
import Selection

from parser_functions import get_basedict, update_basedict, torsion, virtual_torsion, valence, NullValue, get_SPATH, norm, switch_hbonds
from ParserException import ParserException

STD_BASE = ('A', 'ADE', 'C', 'CYT', 'G', 'GUA', 'T', 'THY', 'U', 'URA')

class StructureBuilder:
    """
    Constructing the Structure object and other objects contained in a structure.
    The PDBParser class calls this class.
    """
    def __init__(self):
        self.line_counter = 0
        self.header = {}

    # Private methods
    def _calc_torsions(self, chain):
        """
        Procedure:
        1. Sort residues
        2. Get the target residue, and then its neighbors.        
        Some torsion angles need residue(i-1, i) or residue(i, i+1), e.g.
        alpha in nucleic acid structures.
        3. Check chemical identity of the residue: A (amino acid) or Y/R
        (nucleic acid), then get coordinates.
        4. Check the linkage between three neighbor residues if available.        

        Backbond atom list of nucleic acid: P, O5', C5', C4', C3', O3'
          o alpha:   O3'(i-1)-P-O5'-C5'
          o beta:    P-O5'-C5'-C4'
          o gamma:   O5'-C5'-C4'-C3'
          o delta:   C5'-C4'-C3'-O3'
          o epsilon: C4'-C3'-O3'-P(i+1)
          o zeta:    C3'-O3'-P(i+1)-O5'(i+1)
        chi angle atom list:
          o Purine:      O4'-C1'-N9-C4
          o Pyrimidine:  O4'-C1'-N1-C2
          o P (special): O4'-C1'-C5-C4
        """
        torsion_atoms = ("P", "O5'", "C5'", "C4'", "C3'", "O3'")
        chain.sort()
        res_list = chain.get_child_list()
        for i in range(0, len(res_list)):
            res = res_list[i]
            if len(res_list) == 1:  # There is only ONE residue in a given chain
                res_left = None
                res_right = None
            elif i == 0:
                res_left = None
                res_right = res_list[i+1]
            elif i == len(res_list) - 1:
                res_left = res_list[i-1]
                res_right = None
            else:
                res_left = res_list[i-1]
                res_right = res_list[i+1]
            if res.get_ident() == 'Y' or res.get_ident() == 'R':
                coord_list = []   # list of 9 arrays
                if res_left != None:
                    coord_list.append(res_left.find_atom_coord("O3'"))
                else:
                    coord_list.append(None)
                for atom in torsion_atoms:
                    coord_list.append(res.find_atom_coord(atom))
                if res_right != None:
                    coord_list.append(res_right.find_atom_coord("P"))
                    coord_list.append(res_right.find_atom_coord("O5'"))
                else:
                    coord_list.append(None)
                    coord_list.append(None)
                # Calculate torsion angles
                # six continuous angles and one chi angle
                torsion_array = zeros((7,), 'f') + NullValue
                for j in range(0, 6):
                    v1 = coord_list[j]
                    v2 = coord_list[j+1]
                    v3 = coord_list[j+2]
                    v4 = coord_list[j+3]
                    if v1 is not None and v2 is not None and v3 is not None and v4 is not None:
                        tor_angle = torsion(v1, v2, v3, v4)
                        if tor_angle == None:
                            #sys.stderr.write("of %s\n" % res)
                            tor_angle = NullValue
                        torsion_array[j] = tor_angle
                # Determine the chi angle
                if res.get_ident() == 'R':
                    v1 = res.find_atom_coord("O4'")
                    v2 = res.find_atom_coord("C1'")
                    v3 = res.find_atom_coord("N9")
                    v4 = res.find_atom_coord("C4")
                else:
                    v1 = res.find_atom_coord("O4'")
                    v2 = res.find_atom_coord("C1'")
                    if res.get_refname() == "P":
                        v3 = res.find_atom_coord("C5")
                        v4 = res.find_atom_coord("C4")
                    else:
                        v3 = res.find_atom_coord("N1")
                        v4 = res.find_atom_coord("C2")
                if v1 is not None and v2 is not None and v3 is not None and v4 is not None:
                    tor_angle = torsion(v1, v2, v3, v4)
                    if tor_angle == None:
                        #sys.stderr.write("of %s\n" % res)
                        tor_angle = NullValue
                    torsion_array[6] = tor_angle
                res.set_torsion(torsion_array)
            elif res.get_ident() == 'A':
                # Apply later if needed
                continue
            else:
                continue
            
    def _identity(self, residue):
        """
        Identify chemical identity of the residue.
        Return value:
          R - purine
          Y - pyrimidine
          A - amino acid
          O - other (including water, ligand, small melecule)
        """
        id = "O"
        N9 = residue.find_atom("N9")
        C8 = residue.find_atom("C8")
        N1 = residue.find_atom("N1")
        C2 = residue.find_atom("C2")
        C6 = residue.find_atom("C6")
        if N1 and C2 and C6:
            d1 = C2 - N1
            d2 = C6 - N1
            d3 = C6 - C2
            if d1 <= 2.0 and d2 <= 2.0 and d3 <= 3.0:
                id = "Y"
                if N9:
                    d3 = N9 - N1
                    if d3 >= 3.5 and d3 <= 4.5:
                        id = "R"
                elif C8:
                    d3 = C8 - N1
                    if d3 >= 3.9 and d3 <= 4.9:
                        id = "R"
                return id
        CA = residue.find_atom("CA")
        C = residue.find_atom("C")
        if not C:   # if C does not exist, use N
            C = residue.find_atom("N")
        if CA and C:
            d4 = CA - C
            if d4 >= 0 and d4 <= 2.0:
                id = "A"
        return id

    def _base_ref(self, residue):
        """
        Identify standard base name.
        o residue: residue instance
        o Return values:
        refname - standard base name
        """
        # Automatic check
        ident = residue.get_ident()
        if ident == "R":
            N2 = residue.find_atom("N2")
            if N2:
                refname = "G"
            else:
                refname = "A"
        elif ident == "Y":
            C5M = residue.find_atom("C5M")
            N4 = residue.find_atom("N4")
            O4 = residue.find_atom("O4")
            O2p = residue.find_atom("O2'")
            if (not O2p) and (C5M or (C5M and O4)):
                refname = "T"
            else:
                if N4:
                    refname = "C"
                else:
                    refname = "U"
        else:
            print "Warning: not base <_base_ref>"
            refname = None
        return refname

    def _standard_base(self, resname):
        """
        Check base modification
        Arguments:
        o resname - residue name without whitespaces on both ends
        Return values:
        o modification - string, 'standard'/'modified'
        """
        resname = upper(resname)
        for stdname in STD_BASE:
            if resname == stdname:
                return "standard"
        return "modified"

    def _sugar_para(self, residue):
        """
        Set the sugar conformational parameters of nucleotide.
        o v0: C4'-O4'-C1'-C2'
        o v1: O4'-C1'-C2'-C3'
        o v2: C1'-C2'-C3'-C4'
        o v3: C2'-C3'-C4'-O4'
        o v4: C3'-C4'-O4'-C1'
        o tm: amplitude of pseudorotation of the sugar ring
        o P:  phase angle of pseudorotation of the sugar ring
        o sugar puckering
          10 puckering modes are defined in terms of the phase angle of
          pseudorotation (reference: The Principle of Nucleic Acid Structure,
          Saenger, p19)
          e.g., C3'-endo has P (0, 36)
        """
        atoms = ("C4'", "O4'", "C1'", "C2'", "C3'", "C4'", "O4'", "C1'")
        sugar_array = zeros((7,), 'f') + NullValue
        for i in range(0, 5):
            try:
                va = residue[atoms[i]].get_coord()
                vb = residue[atoms[i+1]].get_coord()
                vc = residue[atoms[i+2]].get_coord()
                vd = residue[atoms[i+3]].get_coord()
            except KeyError:
                print "Some sugar atoms missing in %s" % residue
                return sugar_array, " "
            vi = torsion(va, vb, vc, vd)
            if vi == None:
                #sys.stderr.write("of %s\n" % res)
                return sugar_array, " "
            sugar_array[i] = vi
        P = atan2((sugar_array[4] + sugar_array[1]) - (sugar_array[3] + sugar_array[0]), 2 * sugar_array[2] * (sin(pi/5) + sin(pi/2.5)))
        tm = sugar_array[2] / cos(P)
        # Convert P angle to degree, and P -> (0, 360)
        P = degrees(P)
        if P < 0:
            P = P + 360.0
        sugar_array[5] = tm
        sugar_array[6] = P
        # Determine sugar puckering
        sugar_pucker = ("C3'-endo", "C4'-exo ", "O4'-endo", "C1'-exo ", "C2'-endo",
                        "C3'-exo ", "C4'-endo", "O4'-exo ", "C1'-endo", "C2'-exo ")
        m = int(floor(P / 36.0))
        puckering = sugar_pucker[m]
        return sugar_array, puckering


    # Public methods
    # Initiation methods
    def init_structure(self, structure_id):
        """
        Initiate a new Structure object with given id.
        Arguments:
        o structure_id - string, usually is the NDB id
        """
        self.structure = Structure(structure_id)

    def init_model(self, model_id):
        """
        Initiate a new Model object with given id.
        Arguments:
        o id - int, model id
        """
        self.model = Model(model_id)
        self.structure.add(self.model)

    def init_chain(self, chain_id):
        """
        Initiate a new Chain object with given id.
        Arguments:
        o chain_id - string
        """
        if self.model.has_id(chain_id):
            self.chain = self.model[chain_id]
            if __debug__:
                print "WARNING: Chain %s is discontinuous at line %d." % (chain_id, self.line_counter)
        else:
            self.chain = Chain(chain_id)
            self.model.add(self.chain)

    def init_residue(self, res_id, counter, resseq, icode, resname, hetero_flag, seg_id):
        """
        Initiate a new Residue object.
        Arguments:
        o res_id - tuple, identifier of the residue
        o resseq - int, sequence identifier
        o icode - character, insertion code
        o resname - string, e.g., 'G'
        o hetero_flag - hetero flag, 'W' for water, 'H' for
            hetero residues, otherwise a whitespace
        o seg_id - string
        Procedure:
        1. Check if the residue exits
        2. If it exits, define the current residue as the existing one
        3. If it is a new residue, create a new object and add it to the current chain
        """
        if self.chain.has_id(res_id):
            # There already is a residue with the give residue_id.
            self.residue = self.chain[res_id]
            print "ParserException: Residue (%d, '%s') is redefined" % (resseq, icode)
        else:
            self.residue = Residue(res_id, counter, resseq, icode, resname, hetero_flag, seg_id)
            self.chain.add(self.residue)

    def init_atom(self, aname, counter, serial_number, fullname, coord, bfactor, occupancy, altloc):
        """
        Initate a new atom object.
        Arguments:
        o aname - identifier of the atom
        o counter - a counter for atom numbers
        o serial_number - int
        o fullname - name with spaces if available
        o coord - array
        o bfactor - float
        o occupancy - float
        o altloc - char
        Procedure:
        1. Check if this atom is already present in the residue.
        2. If it is, it might be due to the fact that the two atoms
           have atom names that differ only in spaces (e.g. 'CA..' and '.CA.',
           where the dots are spaces). If that is so, use all spaces
           in the atom name of the current atom. 
        3. The atom is disordered and is present in the current residue.
           Get the disordered atom wrapper, and add a new disordered atom.
        4. The atom is disordered and is not present in the current residue.
           Create a disordered atom wrapper, and add the disordered atom.
        5. New atom and not disordered.
           Create a new atom, and add to the current residue.
        """
        if self.residue.has_id(aname):
            # different in fullname
            duplicated_atom = self.residue[aname]
            duplicate_fullname = duplicated_atom.get_fullname()
            if fullname != duplicate_fullname:
                # use fullname as the atom id
                aname = fullname
                if __debug__:
                    print "WARNING: atom names %s and %s differ only in spaces at line %d." % (duplicate_fullname, fullname, self.line_counter)
        atom = Atom(aname, counter, serial_number, fullname, coord, bfactor, occupancy, altloc)
        self.atom = atom
        if altloc != " ":
            # The atom is disordered
            if self.residue.has_id(aname):
                # The current residue already contains the disordered atom
                duplicated_atom = self.residue[aname]
                if duplicated_atom.is_disordered() == 2:
                    duplicated_atom.disordered_add(atom)
                else:
                    # This is an error in the PDB file:
                    # a disordered atom is found with a blank altloc
                    # Detach the duplicate atom, and put it in a 
                    # DisorderedAtom object together with the current atom.
                    self.residue.detach_child(aname)
                    disordered_atom = DisorderedAtom(aname)
                    self.residue.add(disordered_atom)
                    disordered_atom.disordered_add(atom)
                    disordered_atom.disordered_add(duplicated_atom)
                    if __debug__:
                        print "WARNING: disordered atom found with blanc altloc before line %d." % self.line_counter
            else:
                # The current residue does not contain the disordered atom.
                # Create a new disordered atom wrapper
                disordered_atom = DisorderedAtom(aname)
                self.residue.add(disordered_atom)
                # Add the real atom to the disordered atom wrapper
                disordered_atom.disordered_add(atom)
        else:
            # The atom is not disordered
            self.residue.add(atom)

    # Set methods
    def set_header(self, header):
        "Set the PDB headet dict to the structure instance."
        self.structure.set_header(header)

    def set_line_counter(self, line_counter):
        """
        The line counter keeps track of the line in the PDB file that
        is being parsed.

        Arguments:
        o line_counter - int
        """
        self.line_counter = line_counter

    def set_anisou(self, anisou_array):
        "Set anisotropic B factor of the current atom."
        self.atom.set_anisou(anisou_array)

    def set_siguij(self, siguij_array):
        "Set standard deviation of anisotropic B factor of the current atom."
        self.atom.set_siguij(siguij_array)

    def set_sigatm(self, sigatm_array):
        "Set standard deviation of atom position of the current atom."
        self.atom.set_sigatm(sigatm_array)

    def set_res_attrbt(self):
        """
        For all kinds of residues:
        o ident - A/Y/R/O
        For nucleotides:
        o refname - standard base name, otherwise return " "
        o modification - modification of base
        o sugar parameters

        Set the chemical identity and the reference base name (if available)
        of the residue.
        1. set the chemical identity
        2. if it is a nucleotide, set the reference base name (A/G/U/C/T)
        3. report NEW special base (not in the file 'baselist.dat')

        """
        # Get modified:standard base dict
        basedict = get_basedict()
        newbase = {}  # {resname: (refname, ndb_id)}
        for model in self.structure.get_child_list():
            for chain in model.get_child_list():
                for residue in chain.get_child_list():
                    # chemical identity
                    ident = self._identity(residue)
                    residue.set_ident(ident)
                    # aumatically check the standard base
                    if ident == "R" or ident == "Y":
                        resname = residue.get_resname()
                        if resname in basedict.keys():
                            refname = basedict[resname]
                        else:
                            # new modified base
                            refname = self._base_ref(residue)
                            newbase[resname] = (refname, self.structure.get_id())
                        residue.set_refname(refname)
                        # modification of base
                        modification = self._standard_base(resname)
                        residue.set_modification(modification)
                        # sugar parameters
                        sugar_array, puckering = self._sugar_para(residue)
                        residue.set_sugar_para(sugar_array)
                        residue.set_puckering(puckering)
        # Write the new 'baselist.dat' file if any new nucleotides are found
        if newbase:
            update_basedict(newbase)
                    
    def set_angles(self):
        "Set torsion angles of residues of the current structure."
        for model in self.structure.get_child_list():
            for chain in model.get_child_list():
                self._calc_torsions(chain)

    
    # Get methods
    def get_structure(self):
        "Return the parsed Structure object."
        return self.structure
    

class X3DNABuilder:
    """
    Constructing the x3dna object and other objects contained in an x3dna 
    parser object.
    The X3DNAParser class calls this class.
    """
    def __init__(self):
        pass

    # Private methods
    def _get_hb_pat(self, hbonds):
        """
        Return hydrogen bonding pattern (without distance).
        Arguments:
        o hbonds - (aname1, aname2, hb_dist, hb_attrbt)
        Return values:
        o hb_pat - sorted hbond pattern, (aname1, aname2, hb_attrbt), tuple
        """
        hb_pat = []
        for hb in hbonds:
            pattern = (hb[0], hb[1], hb[3])
            hb_pat.append(pattern)
        hb_pat.sort()
        return tuple(hb_pat)

    def _det_bp_type(self, refname1, refname2, hbonds):
        """
        Determine base-pair class.
        Arguments:
        o refname1 - reference base name of residue1
        o refname2 - reference base name of residue2
        o hbonds - hydrogen bonds
        Return values:
        o bp_type - 
          primary feature - single/regular
          secondary features - bifurcated, symmetric
        """
        # Get unique atom list of two bases
        seen1 = {}
        seen2 = {}
        atoms1 = []
        atoms2 = []
        unique_atoms1 = []
        unique_atoms2 = []
        for hb in hbonds:
            atoms1.append(hb[0])
            atoms2.append(hb[1])
            if not seen1.has_key(hb[0]):   # for base1
                unique_atoms1.append(hb[0])
                seen1[hb[0]] = 1
            if not seen2.has_key(hb[1]):   # for base2
                unique_atoms2.append(hb[1])
                seen2[hb[1]] = 1
        atoms1.sort()
        atoms2.sort()
        unique_atoms1.sort()
        unique_atoms2.sort()
        if len(hbonds) == 1:
            bp_type = ["single"]
        elif len(unique_atoms1) == 1 or len(unique_atoms2) == 1:
            bp_type = ["bifurcated"]
        else:
            bp_type = ["regular"]
            if len(atoms1) > len(unique_atoms1) or len(atoms2) > len(unique_atoms2):
                bp_type.append("bifurcated")
            if refname1 == refname2 and atoms1 == atoms2:
                bp_type.append("symmetric")                
            #if len(unique_atoms1) == len(unique_atoms2):
            #    if refname1 == refname2 and atoms1 == atoms2:
            #        bp_type.append("symmetric")
            #else:
            #    bp_type.append("bifurcated")
            #    if refname1 == refname2 and self._check_sym(unique_atoms1, unique_atoms2):
            #        bp_type.append("symmetric")
        bp_type = tuple(bp_type)   # convert to tuple
        return bp_type

    def _check_sym(self, unique_atoms1, unique_atoms2):
        """
        Arguments:
        o unique_atoms - sorted list
          two lists have different length
        Return values:
        o boolean - symmetric or not
        """
        if len(unique_atoms1) < len(unique_atoms2):
            for atom in unique_atoms1:
                if atom not in unique_atoms2:
                    return False
        else:
            for atom in unique_atoms2:
                if atom not in unique_atoms1:
                    return False
        return True

    def _get_order(self, residue1, residue2, hbonds, bp_para, bp_orie):
        """
        Reorder two bases according to base identities, hydrogen bonds,
        or base-pair parameters.
        
        Arguments:
        o residue1 - residue object
        o residue2 - residue object
        o hbonds - hydroge bonds, tuple
        o bp_para - base-pair parameters, array
        o bp_orie - orientation of z-axis

        Return values:
        o bp_order - 1/2/-1, base order
          1 - use original order
          2 - switch two bases
          -1 - specail base pair, cannot define the order, keep its
               original order, but label the base pair as 'special'
        
        Procedure
        1. base identities
        2. hydrogen bonds
        3. base-pair parameters
        """
        base_to_order = {'G':1, 'A':2, 'C':3, 'U':4, 'T':5, 'I':6, 'P':7}
        base_ident1 = residue1.get_ident()
        base_ident2 = residue2.get_ident()
        if base_ident1 == 'A' or base_ident1 == 'O' or base_ident2 == 'A' or base_ident2 == 'O':
            print "ParserException: Residue is not a nucleotide"
        # Get atom lists
        anames1, anames2 = self._get_anames(hbonds)
        # Reorder procedure:
        # 1. Compare bases
        base_order1 = base_to_order[upper(residue1.get_refname())]
        base_order2 = base_to_order[upper(residue2.get_refname())]
        if base_order1 < base_order2:
            return 1
        elif base_order1 > base_order2:
            return 2
        # 2. Same bases, compare hydrogen bonds
        anames1.sort()
        anames2.sort()
        for i in range(0, len(anames1)):
            if anames1[i] < anames2[i]:
                return 1
            elif anames1[i] > anames2[i]:
                return 2
        # 3. Same bases, and symetric hydrogen bonds, compare Shear and/or Opening
        if abs(bp_para[0]) > 1:   # Shear for anti-parallel or paralle
            if bp_para[0] > 1:
                return 1
            else:
                return 2
        if abs(bp_para[5]) > 50 and bp_orie == 'parallel':   # Opening only for parallel
            if bp_para[5] > 50:
                return 1
            else:
                return 2
        # 4. All other cases, return -1 for bp_order
        return -1

    def _switch_hbonds(self, hbonds):
        """
        Switch the order of atoms in hbonds.
        Arguments:
        o hbonds - tuple
        Return values:
        o hbonds - reordered, tuple
        """
        hbonds_new = []
        for hb in hbonds:
            hb_new = (hb[1], hb[0], hb[2], hb[3])
            hbonds_new.append(hb_new)
        return tuple(hbonds_new)        

    def _get_anames(self, hbonds):
        """
        Return two atom lists of hydrogen bonds.
        Arguments:
        o hbonds - hydrogen bonds, tuple of tuple
        Return values:
        o atoms1 - list of atom names of atom1
        o atoms2 - list of atom namee of atom2
        """
        atoms1 = []
        atoms2 = []
        for hb in hbonds:
            atoms1.append(hb[0])
            atoms2.append(hb[1])
        return atoms1, atoms2

    def _calc_iso_para(self, residue1, residue2):
        """
        Calculate isosteric parameters of the base pair.

        Arguments:
        o residue1 - the residue object
        o residue2 - the residue object

        Return values:
        o iso_para - array of isosteric parameters
          o lambda1: valence angle of residue1, N1/N9-C1'-C1'
          o lambda2: valence angle of residue2, N1/N9-C1'-C1'
          o C1C1: distance, C1'-C1'
          o RN9_YN1: distance, RY9-YN1
          o RC8_YC6: distance, RC8-YC6
          o tor_C1C1: torsion angle, N1/N9-C1'-C1'-N1/N9
        """
        coord_list1 = self._get_iso_coord(residue1)
        coord_list2 = self._get_iso_coord(residue2)
        if coord_list1 == None or coord_list2 == None:
            iso_para = zeros((6,), 'f') + NullValue
            return iso_para
        lambda1 = valence(coord_list1[1], coord_list1[0], coord_list2[0])
        lambda2 = valence(coord_list2[1], coord_list2[0], coord_list1[0])
        C1C1 = norm(coord_list1[0] - coord_list2[0])
        RN9_YN1 = norm(coord_list1[1] - coord_list2[1])
        RC8_YC6 = norm(coord_list1[2] - coord_list2[2])
        tor_C1C1 = virtual_torsion(coord_list1[1], coord_list1[0], coord_list2[0], coord_list2[1])
        if lambda1 == None:
            lambda1 = NullValue
        if lambda2 == None:
            lambda2 = NullValue
        if tor_C1C1 == None:
            tor_C1C1 = NullValue
        return array([lambda1, lambda2, C1C1, RN9_YN1, RC8_YC6, tor_C1C1], 'f')

    def _torsion_to_score(self, aname, v1, v2, v3, v4):
        """
        Calculate the virtual torsion angle and determine the corresponding
        scores.
        """
        angle = virtual_torsion(v1, v2, v3, v4)
        if aname == "N2" or aname == "O2":
            if -90.0 <= angle <= 90.0 or angle == None:
                score_list = array([1, 0, 0, 0])  # 'W'
            else:
                score_list = array([0, 0, 1, 0])  # 'S'
        elif aname == "N6" or aname == "O6" or aname == "N4" or aname == "O4":
            if -90.0 <= angle <= 90.0 or angle == None:
                score_list = array([1, 0, 0, 0])  # 'W'
            else:
                score_list = array([0, 1, 0, 0])  # 'H'
        elif aname == "O2'":
            if -90.0 <= angle <= 90.0 or angle == None:
                score_list = array([0, 0, 1, 0])  # 'S'
            else:
                score_list = array([0, 0, 0, 1]) # 'B'
        return score_list

    def _get_iso_coord(self, residue):
        """
        Get coordinate list for atoms involved in isosteric parameters.

        Arguments:
        o residue - residue object

        Return values:
        o coord_list - array list

        Atoms: C1', N1/N9, C6/C8
        Exception: if the base is P, atoms are C1', C5, C6
        """
        coord_list = []
        base_ident = residue.get_ident()
        # get coordinate of C1'
        try:
            C1p = residue["C1'"]
        except KeyError:
            print "ParserException: No atom C1' in the %s" % residue
            return None        
        coord_list.append(C1p.get_coord())
        # get coordinate of N1/N9 and C6/C8
        if base_ident == 'R':
            try:
                N1N9 = residue["N9"]
                C6C8 = residue["C8"]
            except KeyError:
                print "ParserException: No atom C1' in the %s" % residue
                return None
        else:   # 'Y'
            if residue.get_refname() != "P":
                try:
                    N1N9 = residue["N1"]
                    C6C8 = residue["C6"]
                except KeyError:
                    print "ParserException: No atom C1' in the %s" % residue
                    return None
            else:   # 'P'
                try:
                    N1N9 = residue["C5"]
                    C6C8 = residue["C6"]
                except KeyError:
                    print "ParserException: No atom C1' in the %s" % residue
                    return None
        coord_list.append(N1N9.get_coord())
        coord_list.append(C6C8.get_coord())
        return coord_list

    def _chk_unpaired(self, ss_left, ss_right):
        """
        Check the neighbours of a given single-stranded segment.
        Procedure:
        1. find the closest residue neighbour (usually -1/+1 on residue sequence)
        2. get the sec_id and chain_no
        3. check the two sec_id's and chain_no's (left and right)
           case 1: S_b, internal loop
           case 2-1: S_h, terminal loop
           case 2-2: S_hq, like terminal loop, but at the end of quasi-continuous helix
           case 3: S_l, linker between two helices
           case 4: S_e, the end single-stranded segment
        """
        (res_left, res_right) = self._find_neighbours(ss_left, ss_right)
        if res_left == None or res_right == None:
            return 'S_e'
        secid_left = res_left.get_sec_id()
        chainno_left = res_left.get_chainno()
        helixseq_left = res_left.get_helixseq()
        secid_right = res_right.get_sec_id()
        chainno_right = res_right.get_chainno()
        helixseq_right = res_right.get_helixseq()
        if secid_left == secid_right:
            if chainno_left == chainno_right:
                return 'S_b'
            else:
                if helixseq_left == helixseq_right:
                    return 'S_h'
                else:
                    return 'S_hq'
        else:
            return 'S_l'
        
    def _chk_chain(self, hchain):
        """
        Check the continuity of a chain in a given helix.
        Arguments:
        o hchain - the list of residues in one chain of a helix
        
        There are four kinds of situations:
          1. H: in the middle of a continuous helical region
          2. H_e: at the end of a helix and continuous
          3. H_ie: at the end of a helix and being a break
          4. H_qe: at the break of a continuous helical region in a helix
          5. H_i: a break in the helix, i.e., an inserted base pair
          6. H_l: lone base pair of the helix
        The above four values are assigned to the 'continuity' attribute of residues.
        """
        for i in range(0, len(hchain)):
            res = hchain[i]
            resseq = res.get_resseq()
            if len(hchain) == 1:  # The helix only has a lone base pair
                res.set_continuity('H_l')
            elif i == 0:
                resseq_right = hchain[i+1].get_resseq()
                if abs(resseq - resseq_right) == 1:
                    res.set_continuity('H_e')
                else:
                    res.set_continuity('H_ie')
            elif i == len(hchain) - 1:
                resseq_left = hchain[i-1].get_resseq()
                if abs(resseq - resseq_left) == 1:
                    res.set_continuity('H_e')
                else:
                    res.set_continuity('H_ie')
            else:
                resseq_left = hchain[i-1].get_resseq()
                resseq_right = hchain[i+1].get_resseq()
                if abs(resseq - resseq_left) == 1 and abs(resseq - resseq_right) == 1:
                    res.set_continuity('H')
                elif abs(resseq - resseq_left) != 1 and abs(resseq - resseq_right) == 1:
                    res.set_continuity('H_qe')
                elif abs(resseq - resseq_left) == 1 and abs(resseq - resseq_right) != 1:
                    res.set_continuity('H_qe')
                else:
                    res.set_continuity('H_i')

    def _find_neighbours(self, ss_left, ss_right):
        """
        Find the closest residue neighbour 
        (usually -1/+1 on residue sequence in terms of the leftmost and 
        rightmost residues, respectively)
        """
        chain = ss_left.get_parent()
        res_list = chain.get_child_list()
        res_num = len(res_list)
        index_left = res_list.index(ss_left)
        index_right = res_list.index(ss_right)
        if index_left == 0:
            res_left = None
        else:
            res_left = res_list[index_left-1]
        if index_right == res_num - 1:
            res_right = None
        else:
            res_right = res_list[index_right+1]
        return res_left, res_right
        
    # Public methods
    def init_x3dna(self, id):
        """
        Initiate a new X3DNA object with given id (usually NDB id).
        Arguments:
        o id - string, usually NDB id
        """
        self.x3dna = X3DNAInformer(id)

    def init_secondary(self, id):
        """
        Initiate a new Secondary object with given id.
        There is only one secondary object in the x3dna, which contains
        all secondary elements found in the structure by the 3DNA program.
        Arguments:
        o id - label, string, use 'secondary'
        """
        self.secondary = Secondary(id)
        self.x3dna.add(self.secondary)

    def init_tertiary(self, id):
        """
        Initiate a new tertiary object.
        There is only one tertiary object in the x3dna, which contains
        all base pairs not in helices.
        Arguments:
        o id - label, string, use 'tertiary'
        """
        self.tertiary = Tertiary(id)
        self.x3dna.add(self.tertiary)

    def init_stepwrapper(self, id):
        """
        Initiate a new step wrapper.
        There is only one step wrapper in x3dna, which contains helices.
        Arguments:
        o id - label, string, use 'stepwrapper'
        """
        self.stepwrapper = StepWrapper(id)
        self.x3dna.add(self.stepwrapper)

    def init_multiwrapper(self, id):
        """
        Initiate a new multiplet wrapper.
        There is only one multiplet wrapper in x3dna, which contains multiplets.
        Arguments:
        o id - label, string, use 'multiwrapper'
        """
        self.multiwrapper = MultiWrapper(id)
        self.x3dna.add(self.multiwrapper)

    def init_helix(self, sec_id, sec_class):
        """
        Initiate a new secondary helix object.
        Arguments:
        o sec_id - secondary id, int
        o sec_class - the class of secondary element, helix/unpaired, string
        """
        self.helix = Helix(sec_id, sec_class)
        self.secondary.add(self.helix)

    def init_unpaired(self, sec_id, sec_class):
        """
        Initiate a new secondary unpaired object.
        Arguments:
        o sec_id - secondary id, int
        o sec_class - the class of secondary element, helix/unpaired, string
        """
        self.unpaired = Unpaired(sec_id, sec_class)
        self.secondary.add(self.unpaired)

    def init_basepairh(self, hid, residue1, residue2, bp_para, hbonds, bp_orie, helix_seq):
        """
        Initiate a new basepairh object. BasepairH is a subclass of Basepair,
        which is embedded in a helix.
        Arguments:
        o hid - the pair of residue counters with order found in helix, tuple of int
        o residue - residue object, with order found in helix
        o bp_para - base-pair parameters, array
        o hbonds - hydrogen bonds, tuple of tuple
        o bp_orie - orientation of z-axis, string
        o helix_seq - the sequence number of the base pair in helix
        
        Procedure:
        (1.) Determine the interacting edges of the base pair
        1. Change the base-pair variables according to the bp_order
           bp_order = 1, no change
           bp_order = 2, switch residues and hbonds, change sign of bp_para
           bp_order = -1, no change, but label 'special' to the base pair
        2. Calculate the isosteric parameters for the 'reordered' base pair
        3. Get the bp_id for the 'reordered' base pair
           bp_id - the pair of residue counters
        4. Get pat_id
        5. Determine bp_type
        6. Initiate the BasepairH object
        7. Make association with involved residues, i.e., set sec_id to residues
        """
        bp_order = self._get_order(residue1, residue2, hbonds, bp_para, bp_orie)
        if bp_order == 1:
            bp_label = 'original'
        elif bp_order == 2:
            bp_label = 'reversed'
            # 1. switch residues
            tmp = residue1
            residue1 = residue2
            residue2 = tmp
            # 2. change sign of bp_para
            if bp_orie == 'anti-parallel':
                # change sign of 'Shear' and 'Buckle'
                bp_para[0] = -1 * bp_para[0]
                bp_para[3] = -1 * bp_para[3]
            else:
                # parallel, change sign of all parameters
                bp_para = -1 * bp_para
            # 3. switch hbonds: atom1, atom2
            hbonds = self._switch_hbonds(hbonds)
        else:   # bp_order = -1
            bp_label = 'special'
        iso_para = self._calc_iso_para(residue1, residue2)
        bp_id = (residue1.get_counter(), residue2.get_counter())
        hbond_num = len(hbonds)
        # Get pat_id
        refname1 = residue1.get_refname()
        refname2 = residue2.get_refname()
        hb_pat = self._get_hb_pat(hbonds)
        if hb_pat == ():   # Several base pairs have no hydrogen bonds
            pat_id = None
        else:
            pat_id = (refname1, refname2, bp_orie, hb_pat)
        # Determine base-pair class
        bp_type = self._det_bp_type(refname1, refname2, hbonds)
        # Initiate BasepairH
        self.basepairh = BasepairH(bp_id, residue1, residue2, bp_para, iso_para, \
                                   hbonds, hbond_num, bp_order, bp_label, bp_type, \
                                   pat_id, hid, helix_seq)
        self.helix.add(self.basepairh)
        # Add pdist if phosphate atoms exist
        if residue1.find_atom("P") and residue2.find_atom("P"):
            pdist = residue1["P"] - residue2["P"]
            self.basepairh.set_pdist(pdist)
        # Set secondary class and sec_number of residues
        sec_id = self.helix.get_id()
        residue1.set_sec_class("helix")
        residue1.set_sec_id(sec_id)
        residue2.set_sec_class("helix")
        residue2.set_sec_id(sec_id)

    def add_residue_to_unpaired(self, residue):
        """
        Add residue to the unpiared list.
        """
        self.unpaired.add(residue)
        sec_id = self.unpaired.get_id()
        residue.set_sec_class("unpaired")
        residue.set_sec_id(sec_id)

    def set_sec_attrbt(self):
        """
        Set some secondary attributes.
        Attributes:
        o element_number - the number of child in a secondary element
        """
        for element in self.secondary.get_iterator():
            item_list = element.get_child_list()
            number = len(item_list)
            element.set_number(number)
            if element.get_sec_class() == 'helix':
                (hchain1, hchain2) = element.get_chains()
                self._chk_chain(hchain1)
                self._chk_chain(hchain2)
            else:  # unpaired
                element.sort()
                sslist = element.get_child_list()
                ss_left = sslist[0]
                ss_right = sslist[-1]
                ss_continuity = self._chk_unpaired(ss_left, ss_right)
                element.set_continuity(ss_continuity)

    def set_stepwrapper(self, counter_list, step_list, shelix_list, phos_list, groove_list, form_list):
        """
        Initiate helix for step and base step according to Secondary.
        *** 3DNA count all 'real' or 'not real' base steps
        *** only parse the 'real' base steps in helices
        *** eliminate the 'not real' base steps between helices
        Arguments:
        o counter_list - int list of countering numbers
        o step_list - array list of step parameters
        o shelix_list - array list of local helical parameters
        o phos_list - array list of phosphate tranlational parameters
        o groove_list - array list of groove parameters
        o form_list - string list of step form
        """
        secondary = self.secondary
        secondary.sort()
        idx = 0
        for element in secondary.get_child_list():
            if element.get_sec_class() == "helix":
                sec_id = element.get_id()
                # Initiate SHelix
                bplist = element.get_child_list()
                bp_number = element.get_number()
                if bp_number == 1:  # skip the step for a lone helix
                    idx = idx + 1
                    continue
                self.init_shelix(sec_id)
                for i in range(0, bp_number-1):  # bp_number-1 steps
                    basepair1 = bplist[i]
                    basepair2 = bplist[i+1]
                    step_id = (basepair1.get_id(), basepair2.get_id())
                    # Initiate basestep
                    self.init_basestep(step_id, counter_list[idx], i, basepair1, basepair2, step_list[idx], shelix_list[idx], phos_list[idx], groove_list[idx], form_list[idx])
                    idx = idx + 1
                idx = idx + 1  # skip the step between helices

    def init_shelix(self, id):
        """
        Initiate a new SHelix object.
        The object is similar to the Helix object.
        Arguments:
        o id - helix id, same as element id of helix
        """
        self.shelix = SHelix(id)
        self.stepwrapper.add(self.shelix)

    def init_basestep(self, id, counter, helix_seq, basepair1, basepair2, step_para, helix_para, phos_para, groove_para, step_form):
        """
        Initiate a new Basestep object.
        Arguments:
        o id - pair of basepair ids
        o helix_seq - helix sequence number
        o basepair1 - basepair object
        o basepair2 - basepair object
        o step_para - step parameters
        o helix_para - helical parameters
        o phos_para - phosphate translational parameters
        o groove_para - groove parameters
        o step_form - step form
        """
        self.basestep = Basestep(id, counter, helix_seq, basepair1, basepair2, step_para, helix_para, phos_para, groove_para, step_form)
        self.shelix.add(self.basestep)
        
    def chk_basepairh(self, residue1, residue2, bp_counter, orie_para):
        """
        Set the bp_counter and orientation if they for
        a base pair embedded in helix

        Arguments:

        Return values:
        o boolean - true: in helix; false: not in helix
        
        Procedure:
        1. Get the element object on the basis of sec_id of residues
        2. Find the base pair by try bp_id and reversed bp_id
        3. Set the attribute
        """
        if residue1.get_sec_class() == "helix" and residue2.get_sec_class() == "helix" and residue1.get_sec_id() == residue2.get_sec_id():
            sec_id = residue1.get_sec_id()
            element = self.secondary[sec_id]
            bp_id = (residue1.get_counter(), residue2.get_counter())
            bp_id_swi = (residue2.get_counter(), residue1.get_counter())
            if element.has_id(bp_id):
                basepairh = element[bp_id]
                basepairh.set_counter(bp_counter)
                basepairh.set_orientation(orie_para)
                return True
            elif element.has_id(bp_id_swi):
                basepairh = element[bp_id_swi]
                basepairh.set_counter(bp_counter)
                basepairh.set_orientation(orie_para)
                return True
            else:
                return False
        else:
            return False
        
    def init_basepairu(self, bp_counter, residue1, residue2, bp_para, hbonds, orie_para):
        """
        Initiate a new basepairu object. BasepairU is a subclass of Basepair,
        which represent base pairs not form part of helix.

        Arguments:

        Procedure:
        (1.) Determine the interacting edges of the base pair
        1. Change the base-pair variables according to the bp_order
           bp_order = 1, no change
           bp_order = 2, switch residues and hbonds, change sign of bp_para
           bp_order = -1, no change, but label 'special' to the base pair
        2. Calculate the isosteric parameters for the 'reordered' base pair
        3. Get the bp_id for the 'reordered' base pair
           bp_id - the pair of residue counters
        4. Initiate the BasepairU object
        """
        bp_orie = orie_para[2]
        bp_order = self._get_order(residue1, residue2, hbonds, bp_para, bp_orie)
        if bp_order == 1:
            bp_label = 'original'
        elif bp_order == 2:
            bp_label = 'reversed'
            # 1. switch residues
            tmp = residue1
            residue1 = residue2
            residue2 = tmp
            # 2. change sign of bp_para
            if bp_orie == 'anti-parallel':
                # change sign of 'Shear' and 'Buckle'
                bp_para[0] = -1 * bp_para[0]
                bp_para[3] = -1 * bp_para[0]
            else:
                # parallel, change sign of all parameters
                bp_para = -1 * bp_para
            # 3. switch hbonds: atom1, atom2
            hbonds = self._switch_hbonds(hbonds)
        else:   # bp_order = -1
            bp_label = 'special'
        iso_para = self._calc_iso_para(residue1, residue2)
        bp_id = (residue1.get_counter(), residue2.get_counter())
        hbond_num = len(hbonds)
        refname1 = residue1.get_refname()
        refname2 = residue2.get_refname()
        hb_pat = self._get_hb_pat(hbonds)
        if hb_pat == ():   # Several base pairs have no hydrogen bonds
            pat_id = None
        else:
            pat_id = (refname1, refname2, bp_orie, hb_pat)
        bp_type = self._det_bp_type(refname1, refname2, hbonds)
        self.basepairu = BasepairU(bp_id, bp_counter, residue1, residue2, bp_para, \
                                   iso_para, hbonds, hbond_num, bp_order, bp_label, \
                                   bp_type, pat_id, orie_para)
        self.tertiary.add(self.basepairu)
        # Add pdist if phosphate atoms exist
        if residue1.find_atom("P") and residue2.find_atom("P"):
            pdist = residue1["P"] - residue2["P"]
            self.basepairu.set_pdist(pdist)


    def init_multiplet(self, multi_id, multi_counter, quanlity, res_num, res_list):
        """
        Initiate a new multiplet instance.
        
        """
        self.multiplet = Multiplet(multi_id, multi_counter, quanlity, res_num, res_list)
        self.multiwrapper.add(self.multiplet)
        
    def set_LWpat(self, residue1, residue2, edge1, edge2, orientation):
        """
        1. Set the Leontis-Westhof pattern for the base pair.
        2. Set the interacting edges to the corresponding base pair.
        Arguments:
        o residue1, residue2 - residue instances of LW base pair
          type: object
        o lw_pat - LW pattern, e.g., S/W/trans
          type: string
        """
        bp_list = self.tertiary.get_child_list()
        bp_list = bp_list + Selection.sel_basepairh(self.x3dna)
        for bp in bp_list:
            (r1, r2) = bp.get_respair()
            if r1 is residue1 and r2 is residue2:
                bp.set_edges([edge1, edge2])
                bp.set_lworie(orientation)
                break
            elif r1 is residue2 and r2 is residue1:
                bp.set_edges([edge2, edge1])
                bp.set_lworie(orientation)
                break

    def get_x3dna(self):
        "Return the X3DNAInformer instance."
        return self.x3dna


