# Created: 3/16/2005
# This class is part of the 3DNAParser package.
# 
# The PDBParser class was modified from the Bio.PDB.PDBParser class.
#
########
# The Bio.PDB code is reused and modified for the 3DNAParser package.
#
# Copyright (C) 2002, Thomas Hamelryck (thamelry@vub.ac.be)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.  
########

import re
from string import strip, upper
from numpy import array  # new version of numpy 1.0.3

from Builder import StructureBuilder
from parse_pdb_header import _parse_pdb_header_list
from ParserException import ParserException

__doc__ = "Consumer class that builds a Structure object"

class PDBParser:
    "Parse a PDB file and return a Structure object."
    def __init__(self, structure_builder=None):
        """
        The PDB parser call a number of standard methods in an aggregated
        StructureBuilder object. Normally this object is instanciated by the
        PDBParser object itself, but if the user provides his own StructureBuilder
        object, the latter is used instead.

        Arguments:
        o structure_builder - an optional user implemented StructureBuilder class. 
        """
        if structure_builder != None:
            self.structure_builder = structure_builder
        else:
            self.structure_builder = StructureBuilder()
        self.header = None
        self.trailer = None
        self.line_counter = 0


    # Public methods

    def parse_pdb(self, id, filename):
        """
        Return the structure.
        Procedure:
        1. Parse the PDB file and initiate Structure and its subclasses
        2. Set 'header' attributes for Structure
        Set attributes that cannot not be directly obtained from the PDB file
        3. Set some residue attributes, including 'ident', 'refname', sugar para
        4. Set torsion angles to Residue

        Arguments:
        o id - string, usually is the NDB id
        o filename - name of the PDB file
        """
        # Make a StructureBuilder instance
        self.structure_builder.init_structure(id)
        file = open(filename)
        self._parse(file.readlines())
        file.close()
        self.structure_builder.set_header(self.header)
        self.structure_builder.set_res_attrbt()
        self.structure_builder.set_angles()

    def get_structure(self):
        # Return the Structure instance
        return self.structure_builder.get_structure()


    # Private methods

    def _parse(self, header_coords_trailer):
        "Parse the PDB file."
        # Extract the header; return the rest of the file
        self.header, coords_trailer = self._get_header(header_coords_trailer)
        # Parse the atomic data; return the PDB file trailer
        self.trailer = self._parse_coordinates(coords_trailer)

    def _get_header(self, header_coords_trailer):
        "Get the header dict of the PDB file, return the rest."
        structure_builder = self.structure_builder
        for i in range(0, len(header_coords_trailer)):
            structure_builder.set_line_counter(i+1)
            line = header_coords_trailer[i]
            record_type = line[0:6]
            if (record_type == 'ATOM  ' or record_type == 'HETATM' or record_type == 'MODEL '):
                break
        header = header_coords_trailer[0:i]
        # Return the rest of the coords+trailer for further processing
        self.line_couter = i
        coords_trailer = header_coords_trailer[i:]
        header_dict = _parse_pdb_header_list(header)
        return header_dict, coords_trailer

    def _parse_coordinates(self, coords_trailer):
        "Parse the atomic data in the PDB file."
        local_line_counter = 0
        structure_builder = self.structure_builder
        current_model_id = 1
        current_chain_id = None
        current_residue_id = None
        res_counter = 1
        atom_counter = 0
        structure_builder.init_model(current_model_id)
        for i in range(0, len(coords_trailer)):
            line = coords_trailer[i]
            record_type = line[0:6]
            global_line_counter = self.line_counter + local_line_counter + 1
            structure_builder.set_line_counter(global_line_counter)
            if record_type == 'ATOM  ' or record_type == 'HETATM':
                atom_counter = atom_counter + 1
                serial_number = int(line[6:11])   # integer
                fullname = line[12:16]
                # get rid of whitespace on both ends if any
                # e.g., " C2 " -> "C2", " N B " -> "N B"
                # convert * -> ' for backbond atoms
                aname = strip(fullname)
                aname = self._convert_aname(aname)
                altloc = line[16:17]
                resname = strip(line[17:20])   # delete whitespaces
                chainid = upper(line[21:22])
                resseq = int(line[22:26])   # integer
                icode = line[26:27]
                if record_type == "HETATM": 
                    if resname == "HOH" or resname == "WAT":
                        hetero_flag = "W"   # water
                    else:
                        hetero_flag = "H"   # other hetero atoms
                else:
                    hetero_flag = " "
                # residue_id has exceptions in some PDB files because of 
                # the nonstandard format.
                # e.g., {ar0003: A 12}, {trna03: G 1 and SPM 1}
                residue_id = (resname, resseq, icode)
                # atomic coordinates
                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])
                coord = array([x, y, z], 'f')
                # occupancy and B factor
                occupancy = float(line[54:60])
                bfactor = float(line[60:66])
                seg_id = line[72:76]
                if seg_id == "    ":
                    seg_id = " "
                # initiate chain and residue
                if current_chain_id != chainid:
                    current_chain_id = chainid
                    structure_builder.init_chain(current_chain_id)
                    current_residue_id = residue_id
                    try:
                        structure_builder.init_residue(residue_id, res_counter, resseq, icode, resname, hetero_flag, seg_id)
                    except ParserException, einst:
                        print einst.value
                    else:
                        res_counter = res_counter + 1
                elif current_residue_id != residue_id:
                    current_residue_id = residue_id
                    try:
                        structure_builder.init_residue(residue_id, res_counter, resseq, icode, resname, hetero_flag, seg_id)
                    except ParserException, einst:
                        print einst.value
                    else:
                        res_counter = res_counter + 1
                # initiate atom
                try:
                    structure_builder.init_atom(aname, atom_counter, serial_number, fullname, coord, bfactor, occupancy, altloc)
                except ParserException, einst:
                    print einst.value
            elif record_type == 'ANISOU':
                anisou = map(float, (line[28:35], line[35:42], line[43:49], line[49:56], line[56:63], line[63:70]))
                # U's are scaled by 10000
                anisou_array = (array(anisou, 'f')/10000.0).astype('f')
                structure_builder.set_anisou(anisou_array)
            elif record_type == 'ENDMDL':
                current_model_id = current_model_id + 1
                structure_builder.init_model(current_model_id)
                current_chain_id = None
                current_residue_id = None
            elif record_type == 'END   ' or record_type == 'CONECT':
                self.line_counter = self.line_counter + local_line_counter
                return coords_trailer[local_line_counter:]
            elif record_type == 'SIGUIJ':
                # standard deviation of anisotropic B factor
                siguij = map(float, (line[28:35], line[35:42], line[42:49], line[49:56], line[56:63], line[63:70]))
                # U sigma's are scaled by 10000
                siguij_array = (array(siguij, 'f')/10000.0).astype('f')
                structure_builder.set_siguij(siguij_array)
            elif record_type == 'SIGATM':
                # standard deviation of atomic positions
                sigatm = map(float, (line[30:38], line[38:45], line[46:54], line[54:60], line[60:66]))
                sigatm_array = array(sigatm, 'f')
                structure_builder.set_sigatm(sigatm_array)
            local_line_counter = local_line_counter + 1
        # EOF (not end in END or CONECT)
        self.line_counter = self.line_counter + local_line_counter
        return []
                                             
    def _convert_aname(self, aname):
        "Convert uncommon atom names to standard ones"
        aname = re.sub("\*", "'", aname)  # * -> '
        if aname == "O1'":                # O1' -> O4' 
            aname = "O4'"
        elif aname == "OL":               # OL -> O1P
            aname = "O1P"
        elif aname == "OR":               # OR -> O2P 
            aname = "O2P"
        elif aname == "C5A":              # C5A -> C5M
            aname = "C5M"
        elif aname == "O5T":              # terminal O5'
            aname = "O5'"
        elif aname == "O3T":              # terminal O3'
            aname = "O3'"
        return aname
