# Created: 10/14/2004
# This class is part of the 3DNAParser package

########
# The Bio.PDB code is reused and modified for the 3DNAParser package.
#
# Copyright (C) 2002, Thomas Hamelryck (thamelry@vub.ac.be)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.  
########

from string import join
from numpy import array

from Entity import Entity
from parser_functions import contstr


__doc__ = "Residue class for Structure objects."

class Residue(Entity):
    """
    Residue class is a subclass of Entity. It is a child of
    Chain, and contains Atom class. At the same time, it
    is associated with Multibase class.
       Structure <>-- Model <>-- Chain <>-- Residue <>-- Atom
                                               ^
                                               |
         Unpaired, Basepair, and Multibase-----

    Arguments:
    o id - (resname, resseq, icode)
      type: tuple
    o counter - countering number
      type: int
    o resseq - residue sequence number
      type: int
    o icode - insertion code
      type: char
    o resname - residue name
      type: string
    o hetero_flag - label of hetero residue
      type: char
    o seg_id - segment id
      type: int
      
    Other attributes (set in PDBParser and 3DNAParser):
    o ident - residue identity
      type: char Y/R/A/O
    o refname - base reference name (only for nucleic acid residues)
      type: char
    o modification - base modification, standard/modified
      type: string
    o sec_class - class of secondary element that the residue belongs to
      type: string helix/unpaired
    o sec_id - secondary id
      type: int
    o torsion - torsion angles, default value 500.0 (NullValue)
      type: array (Float)
    o sugar - sugar parameters, default value 500.0 (NullValue)
      type: array (Float)
    o puckering - sugar puckering
      type: string
    o continuity - the continuity of the residue in the corresponding 
      secondary structure:
          1. H: in the middle of a continuous helical region
          2. H_e: at the end of a helix and continuous
          3. H_ie: at the end of a helix and being a break
          4. H_qe: at the break of a continuous helical region in a helix
          5. H_i: a break in the helix, i.e., an inserted base pair
          6. H_l: lone base pair of the helix
          7: S_b, internal loop
          8: S_h, terminal loop
          9: S_hq, like terminal loop, but at the end of quasi-continuous helix
          10:S_l, linker between two helices
          11:S_e, the end single-stranded segment
      type: string
    o chain_no - the chain no in a helix, 1/2
      type: int
    o helix_seq - the sequnce number of base pair in a helix
      type: int
    """
    def __init__(self, id, counter, resseq, icode, resname, hetero_flag, seg_id):
        self.level = "R"
        self.counter = counter
        self.resseq = resseq
        self.icode = icode
        self.resname = resname
        self.hetero_flag = hetero_flag
        self.seg_id = seg_id
        self.refname = None
        self.modification = None
        self.ident = None
        self.sec_class = None
        self.sec_id = None
        self.torsion_array = None
        self.sugar_array = None
        self.puckering = None
        self.continuity = None
        self.chain_no = None
        self.helix_seq = None
        Entity.__init__(self, id)

    # Special methods
    def __repr__(self):
        return "<Residue %s resseq=%s icode=%s>" % (self.resname, self.resseq, self.icode)

    # Private methods
    def _sort(self, a1, a2):
        """
        Sort the Atom instances.
        Atoms are sorted by their serial numbers.
        Arguments:
        o a1, a2 - Atom objects
        """
        serial1 = a1.get_serial_number()
        serial2 = a2.get_serial_number()
        return cmp(serial1, serial2)

    # Public methods
    def sort(self):
        self.child_list.sort(self._sort)
        
    def set_ident(self, identity):
        self.ident = identity

    def set_refname(self, refname):
        self.refname = refname

    def set_modification(self, modification):
        self.modification = modification
        
    def set_hetero_flag(self, hetero_flag):
        self.hetero_flag = hetero_flag

    def set_seg_id(self, seg_id):
        self.seg_id = seg_id

    def set_sec_class(self, sec_class):
        "Set the secondary element name."
        self.sec_class = sec_class

    def set_sec_id(self, sec_id):
        "Set the secondary number."
        self.sec_id = sec_id

    def set_torsion(self, torsion_array):
        "Set torsion anlges."
        self.torsion_array = torsion_array

    def set_sugar_para(self, sugar_array):
        "Set sugar phase angles."
        self.sugar_array = sugar_array

    def set_puckering(self, puckering):
        "Set sugar puckering."
        self.puckering = puckering

    def set_continuity(self, continuity):
        "Set the continuity in the secondary environment."
        self.continuity = continuity
    
    def set_chainno(self, chain_no):
        "Set the chain no of residue in a helix."
        self.chain_no = chain_no
        
    def set_helixseq(self, helix_seq):
        "Set the helix seq number of residue in a helix."
        self.helix_seq = helix_seq
        
    def get_counter(self):
        "Return the residue contering number."
        return self.counter

    def get_resseq(self):
        "Return the residue sequence number."
        return self.resseq

    def get_icode(self):
        "Return the insertion code."
        return self.icode
    
    def get_resname(self):
        "Return the residue name."
        return self.resname

    def get_refname(self):
        "Return the reference name of the residue."
        return self.refname

    def get_modification(self):
        "Return the modification of the nucleotide."
        return self.modification

    def get_hetero_flag(self):
        "Return the hetero flag of the residue."
        return self.hetero_flag

    def get_seg_id(self):
        "Return the seg_id of the residue."
        return self.seg_id

    def get_ident(self):
        "Return the residue identity."
        return self.ident
    
    def get_sec_class(self):
        "Return the secondary element that the residue belongs to."
        return self.sec_class

    def get_sec_id(self):
        "Return the secondary number of the residue."
        return self.sec_id

    def get_torsion(self):
        "Return the torsion angles of the residue (with .1 precision)."
        return self.torsion_array
    
    def get_sugar_para(self):
        "Return the sugar parameters of the residue (with .1 precision)."
        return self.sugar_array
    
    def get_puckering(self):
        "Return the puckering parameters of the residue."
        return self.puckering

    def get_continuity(self):
        return self.continuity
    
    def get_chainno(self):
        return self.chain_no
    
    def get_helixseq(self):
        return self.helix_seq

    def get_coord_array(self):
        "Return coordinate array of the residue."
        atom_list = self.get_child_list()
        coord_list = []
        for atom in atom_list:
            coord_list.append(atom.get_coord())
        return array(coord_list, 'f')

    def get_unpacked_list(self):
        """
        Returns the list of all atoms, unpack DisorderedAtoms.
        """
        atom_list = self.get_child_list()
        undisordered_atom_list = []
        for atom in atom_list:
            if atom.is_disordered():
                undisordered_atom_list = undisordered_atom_list + atom.disordered_get_list()
            else:
                undisordered_atom_list.append(atom)
        return undisordered_atom_list       

    def find_atom(self, aname):
        """
        Find atom by name.
        Return the atom object if it is true, otherwise return None.
        If the atom is disordered, return the selected atom.
        """
        try:
            found_atom = self[aname]
        except KeyError:
            # print "Atom <%s> not found in %s" % (aname, self)            
            found_atom = None
        return found_atom

    def find_atom_coord(self, aname):
        """
        Find atom by name, and get its coordinates.
        Return None if the atom is missing.
        """
        atom = self.find_atom(aname)
        if atom == None:
            return None
        return atom.get_coord()
