# Created: 6/2/2005
# functions used for parsing 3DNA files

########
# The Bio.PDB code is reused and modified for the 3DNAParser package.
#
# Copyright (C) 2002, Thomas Hamelryck (thamelry@vub.ac.be)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.  
########

from types import ListType

from ParserException import ParserException

__doc__ = "Selection of atoms, residues, etc."

pdb_levels = ["A", "R", "C", "M", "S"]

def uniqueify(l):
    "Return unique items in list l."
    d = {}
    for i in l:
        if not d.has_key(i):
            d[i] = None
    return d.keys()

def get_unique_parents(entity_list):
    """
    Translate a list of entities to a list of their
    (unique) parents.
    """ 
    l = []
    for entity in entity_list:
        parent = entity.get_parent()
        l.append(parent)
    return uniqueify(l)

def unfold_entities(entity_list, target_level):
    """
    Return (unpacked) list of entities for PDBParser marked by target level.
    Arguments:
    o entity_list - list of entities or a single entity
      type: list or a single instance
    o target_level - level mark, A/R/C/M/S
      type: string
    Return values:
    o unpacked entity list (or a single entity)
    """
    if not target_level in pdb_levels:
        raise ParserException(target_level + ": not an entity level.")
    if type(entity_list) != ListType:
        # single entity
        entity_list = [entity_list]
    # level of entity list
    level = entity_list[0].get_level()
    for entity in entity_list:
        if not (entity.get_level() == level):
            raise ParserException("Entity list is not homogeneous.")
    target_index = pdb_levels.index(target_level)
    level_index = pdb_levels.index(level)
    if level_index == target_index:
        # already right level
        # return unpacked atom list if the target level is 'A'
        if target_index == 0:
            new_entity_list = []
            for entity in entity_list:
                if entity.is_disordered():
                    new_entity_list = new_entity_list + entity.disordered_get_list()
                else:
                    new_entity_list.append(entity)
            entity_list = new_entity_list            
        return entity_list
    if level_index > target_index:
        # we're going down, e.g. S -> A
        for i in range(target_index, level_index):
            new_entity_list = []
            for entity in entity_list:
                if entity.get_level() == 'R':
                    # unpack disordered atoms if any
                    new_entity_list = new_entity_list + entity.get_unpacked_list()
                else:
                    new_entity_list = new_entity_list + entity.get_child_list()
            entity_list = new_entity_list
    else:
        # we're going up, e.g. A -> S
        for i in range(level_index, target_index):
            new_entity_list = []  
            for entity in entity_list:
                parent = entity.get_parent()
                new_entity_list.append(parent)
            # find unique parents
            entity_list = uniqueify(new_entity_list)
    return entity_list

def sel_basepairh(x3dna):
    """
    Get BasepairH entity list.
    Arguments
    o x3dna - the X3DNA parser
      type: X3DNA instance
    Return values:
    o basepairh list (or a single entity)
    """
    bph_list = []
    secondary = x3dna['secondary']
    for element in secondary.get_iterator():
        if element.get_sec_class() == 'helix':
            bph_list = bph_list + element.get_child_list()
    return bph_list

def sel_basepairu(x3dna):
    "Get BasepairU entity list."
    tertiary = x3dna['tertiary']
    return tertiary.get_child_list()

def sel_basestep(x3dna):
    "Get Basestep entity list."
    step_list = []
    stepwrapper = x3dna['stepwrapper']
    for helix in stepwrapper.get_iterator():
        step_list = step_list + helix.get_child_list()
    return step_list

def sel_multiplet(x3dna):
    "Get Multiplet entity list."
    multiwrapper = x3dna['multiwrapper']
    return multiwrapper.get_child_list()

def sel_secondary(x3dna):
    "Get secondary elements."
    secondary = x3dna['secondary']
    return secondary.get_child_list()
