# Created: 10/12/2004
# This class is part of the 3DNAParser package

########
# The Bio.PDB code is reused and modified for the 3DNAParser package.
#
# Copyright (C) 2002, Thomas Hamelryck (thamelry@vub.ac.be)
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.  
########

from string import join

from Entity import Entity

__doc__ = "The PDB structure class for a macromolecular structure."

class Structure(Entity):
    """
    The hierarchy of Structure is based on SMCRA
    (Structure/Model/Chain/Residue/Atom) scheme. 

    Structure is the highest level of the hierarchy of a molecule.
    The class of Structure contains child list Model.
    Arugments:
    o id - ndb id (string)
    o header - dict of PDB header attributes
        pdb_id - pdb id (string)
        head - HEADER in PDB file (string)
        title - TITLE in PDB file (string)
        resolution - (Float, 2 decimal precision)
        ph - (Float, 2 decimal precision)
        method - synthetic method (string)
        deposotion_date - depostion date (string)
        release_date - release date (string)
    """
    def __init__(self, id):
        self.level = "S"
        self.header = None
        Entity.__init__(self, id)

    # Special methods
    def __repr__(self):
        "Represent Structure as its id (NDB id for default)."
        return "<Structure id = %s>" % self.id

    # Private methods
    def _sort(self, m1, m2):
        """
        Sort models.
        
        This sorting function sorts the Model instances in the Structure
        instance. The sorting is done based on the model id, which is a
        simple int that reflects the order of the models in the PDB file.

        Arguments:
        o m1, m2 - Model instances        
        """
        return cmp(m1.get_id(), m2.get_id())

    # Public method
    def set_header(self, header):
        "Set the header dict."
        self.header = header
    
    def set_pdb_id(self, pdb_id):
        self.header['pdb_id'] = pdb_id

    def set_resolution(self, resl):
        self.header['resolution'] = resl

    def set_ph(self, ph):
        self.header['ph'] = ph

    def set_head(self, head):
        self.header['head'] = head

    def set_title(self, title):
        self.header['title'] = title

    def set_dep_date(self, dep_date):
        self.header['depositon_date'] = dep_date

    def set_rel_date(self, rel_date):
        self.header['release_date'] = rel_date        

    def get_headdict(self):
        "Get the header dict."
        return self.header
    
    def get_pdb_id(self):
        "Return the pdb id otherwise return 'unknown'."
        pdb_id = self.header['pdb_id']
        if pdb_id == '\N':
            pdb_id = 'unknown'
        return pdb_id

    def get_resolution(self):
        "Return the resolution of the structure."
        return self.header['resolution']

    def get_ph(self):
        "Return the ph value of the structure."
        return self.header['ph']

    def get_head(self):
        "Return the header in the PDB file."
        return self.header['head']

    def get_title(self):
        "Return the title in the PDB file."
        return self.header['title']

    def get_dep_date(self):
        "Return the deposition date of the structure."
        return self.header['deposition_date']

    def get_rel_date(self):
        "Return the release date of the structure."
        return self.header['release_date']
    
    def get_source(self):
        """
        Return compound, source entries of the structure.
        Attributes:
        ndb_id mol_id compound source
        Return values:
        o record_list - list of strings (entries)
        """
        ndb_id = self.id
        sid = 1
        record_list = []
        for key in self.header['compound'].keys():
            mol_id = key
            compound = self.header['compound'][key]  # dict, usually {'misc':''}
            for cval in compound.keys():
                if compound[cval] == '':
                    continue
                record = join([ndb_id, str(sid), mol_id, cval, compound[cval]], '\t')
                sid += 1
                record_list.append(record)
            source = self.header['source'][key]  # dict
            for sval in source.keys():
                if source[sval] == '':
                    continue
                record = join([ndb_id, str(sid), mol_id, sval, source[sval]], '\t')
                sid += 1
                record_list.append(record)
        return record_list
    
    def get_residue(self, model_id, chain_id, residue_id):
        """
        Return the residue object in the structure.
        """
        model = self[model_id]
        chain = model[chain_id]
        return chain[residue_id]

    def find_residue(self, model_id, chain_id, refname, res_seq):
        model = self[model_id]
        chain = model[chain_id]
        for residue in chain.get_child_list():
            if res_seq == residue.get_resseq() and refname == residue.get_refname():
                return residue
        return None
