# Created: 6/17/2005
# This class is part of the 3DNA package

from copy import copy
from string import join

from Entity import ChildEntity

__doc__ = "Class for unpaired regions."

class Unpaired(ChildEntity):
    """
    Arguments:
    o id - secondary id (counter)
      type: int
    o sec_class - class of secondary structure, helix/unpaired
      type: string
    o element_number - the number of items
      type: int
    o res_list - list of residue objects
      type: list
    o res_dict - dict of residue objects
      type: dict
    """
    def __init__(self, id, sec_class):
        self.level = 'U'
        self.sec_class = sec_class
        self.element_number = None
        self.res_list = []
        self.res_dict = {}
        ChildEntity.__init__(self, id)

    # Special methods
    def __repr__(self):
        return "<Unpaired region id = %d>" % self.id

    # Private methods
    def _sort(self, r1, r2):
        "Sort resiues in Unpaired region by their countering numbers."
        seq1 = r1.get_counter()
        seq2 = r2.get_counter()
        return cmp(seq1, seq2)

    # Public methods
    def sort(self):
        "Sort residues in a given unpaired region."
        self.res_list.sort(self._sort)
    
    def set_sec_class(self, sec_class):
        "Set secondary class name."
        self.sec_class = sec_class

    def set_number(self, element_number):
        "Set the number of the element."
        self.element_number = element_number

    def set_continuity(self, continuity):
        "Set the attribute of continuity of Unpaired members."
        for residue in self.res_list:
            residue.set_continuity(continuity)

    def add(self, residue):
        "Add residue to the unpaired region."
        res_id = residue.get_id()
        if self.has_id(res_id):
            raise ParserException(str(res_id) + " duplicated")
        self.res_list.append(residue)
        self.res_dict[res_id] = residue

    def has_id(self, id):
        "Return 1 if a residue with given id exists."
        return self.res_dict.has_key(id)

    def get_child_list(self):
        "Return a copy of residue list."
        return copy(self.res_list)
    
    def get_sec_class(self):
        return self.sec_class

    def get_number(self):
        return self.element_number

