# Created: 6/5/2005
# This class is part of the 3DNAParser package.
# 
# The PDBParser class was modified from the Bio.PDB.PDBParser class.

import re
import os
from string import split, upper
from numpy import array, zeros

import Selection
from Builder import X3DNABuilder
from ParserException import ParserException
from parser_functions import NullValue

__doc__ = "Consumer class that build a 3DNA object"

class X3DNAParser:
    def __init__(self, id, structure):
        self.id = id
        self.structure = structure
        self.x3dna_builder = X3DNABuilder()

    # Private methods
    def _get_bpunit_trailer(self, allbp_trailer):
        "Get all bpunit trailer."
        start = -1
        bpunit_trailer = []
        for i in range(0, len(allbp_trailer)):
            if re.search("^   #6: The actual parameters", allbp_trailer[i]):
                start = i + 2
                break
        if start == -1:
            return []
        for i in range(start, len(allbp_trailer), 6):
            line = allbp_trailer[i]
            if re.search("^\s*\d+\s+\d+\s+\d+", line) and re.search("\[", line):
                bpunit_trailer = bpunit_trailer + allbp_trailer[i:i+6]
            else:
                break
        return bpunit_trailer
        
    def _get_multi_trailers(self, allbp_trailer):
        "Get multiplet trailer."
        multiper_trailer = []
        multipar_trailer = []
        marker = 0
        start = -1
        for i in range(0, len(allbp_trailer)):
            line = allbp_trailer[i]
            if re.search("^=+", line) and re.search("perfect match", line) and marker == 0:
                marker = marker + 1
            elif re.search("^=+", line) and re.search("perfect match", line) and marker == 1:
                marker = marker + 1
                start = i + 1
                break
        if start == -1:
            return [], []
        # perfect match
        for i in range(start, len(allbp_trailer)):
            if re.search("^\s*\d+:\s\#\d+\s\[\d+\]", allbp_trailer[i]):
                multiper_trailer.append(allbp_trailer[i])
            else:
                start = i
                break
        # partial match
        for i in range(start, len(allbp_trailer)):
            line = allbp_trailer[i]
            if re.search("^=+", line) and re.search("partial match", line) and marker == 2:
                start = i + 1
        for i in range(start, len(allbp_trailer)):
            if re.search("^\s*\d+:\s\#\d+\s\[\d+\]", allbp_trailer[i]):
                multipar_trailer.append(allbp_trailer[i])
            else:
                break
        return multiper_trailer, multipar_trailer

    def _get_hlxtrailer(self, ana_trailer):
        "Return the helix trailer."
        helix_trailer = []
        start = -1
        for i in range(0, len(ana_trailer)):
            if re.search("^RMSD of the bases", ana_trailer[i]):
                if re.search("^            Strand I", ana_trailer[i+2]):
                    start = i + 3
                    break
        if start == -1:
            return [], ana_trailer
        for i in range(start, len(ana_trailer)):
            if re.search("^\s*\d+\s\s\s\(", ana_trailer[i]):
                helix_trailer.append(ana_trailer[i])
            else:
                break
        return helix_trailer, ana_trailer[i:]

    def _get_hbtrailer(self, trailer):
        "Return the hbond_trailer."
        hbond_trailer = []
        start = -1
        for i in range(0, len(trailer)):
            if re.search("^Detailed H-bond information:", trailer[i]):
                start = i + 1
                break
        if start == -1:
            return [], trailer
        for i in range(start, len(trailer)):
            if re.search("^\s*\d+\s\w-.+-\w\s\s\[\d+\]", trailer[i]):
                hbond_trailer.append(trailer[i])
            else:
                break
        return hbond_trailer, trailer[i:]

    def _get_bptrailer(self, trailer):
        "Return the base-pair parameter trailer."
        bp_trailer = []
        start = -1
        for i in range(0, len(trailer)):
            if re.search("^Local base-pair parameters", trailer[i]):
                start = i + 2
                break
        if start == -1:
            return [], trailer
        for i in range(start, len(trailer)):
            if re.search("^\s*\d+\s\w(\+|-)\w\s+", trailer[i]):
                bp_trailer.append(trailer[i])
            else:
                break
        return bp_trailer, trailer[i:]

    def _get_steptrailer(self, ana_trailer):
        "Return the base step parameter trailer."
        step_trailer = []
        start = -1
        for i in range(0, len(ana_trailer)):
            if re.search("^    step       Shift", ana_trailer[i]):
                start = i + 1
                break
        if start == -1:
            return [], ana_trailer
        for i in range(start, len(ana_trailer)):
            if re.search("^\s*\d+\s\w\w/\w\w\s+", ana_trailer[i]):
                step_trailer.append(ana_trailer[i])
            else:
                break
        return step_trailer, ana_trailer[i:]

    def _get_shlxtrailer(self, trailer):
        "Return the base-pair helical parameter trailer."
        shelix_trailer = []
        start = -1
        for i in range(0, len(trailer)):
            if re.search("^    step       X-disp", trailer[i]):
                start = i + 1
                break
        if start == -1:
            return [], trailer
        for i in range(start, len(trailer)):
            if re.search("^\s*\d+\s\w\w/\w\w\s+", trailer[i]):
                shelix_trailer.append(trailer[i])
            else:
                break
        return shelix_trailer, trailer[i:]

    def _get_phostrailer(self, trailer):
        "Return the phosphate translation trailer."
        phos_trailer = []
        start = -1
        for i in range(0, len(trailer)):
            if re.search("^    step       Xp", trailer[i]):
                start = i + 1
                break
        if start == -1:
            return [], trailer
        for i in range(start, len(trailer)):
            if re.search("^\s*\d+\s\w\w/\w\w\s+", trailer[i]):
                phos_trailer.append(trailer[i])
            else:
                break
        return phos_trailer, trailer[i:]

    def _get_gtrailer(self, trailer):
        "Return the groove parameter trailer."
        groove_trailer = []
        start = -1
        for i in range(0, len(trailer)):
            if re.search("^                 P-P     Refined ", trailer[i]):
                start = i + 1
                break
        if start == -1:
            return [], trailer
        for i in range(start, len(trailer)):
            if re.search("^\s*\d+\s\w\w/\w\w\s+", trailer[i]):
                groove_trailer.append(trailer[i])
            else:
                break
        return groove_trailer, trailer[i:]

    def _parse_hbinfo(self, hb_str):
        """
        Parse hbond information from the string of hbond.

        Arguments:
        o hb_str - string of hbond info
          e.g., [3]  O6 - N4  2.81  N1 - N3  2.92  N2 - O2  2.96
                [11]  O2'- O2P 3.13  O2'- O5' 2.58  O2'- O4' 3.30  ...

        Return values:
        o hbonds - tuple of hbond pattern
        """
        hbonds = []
        tmp_list = hb_str.split("]")
        hbond_num = int(re.sub("\[", "", tmp_list[0]))
        hb_newstr = tmp_list[1]   # "  O6 - N4  2.81  N1 - N3  2.92  N2 - O2  2.96"
        for j in range(0, hbond_num):
            start = 1 + j * 15
            hb_info = hb_newstr[start:start+15]
            # take care of a bug in 3DNA
            if len(hb_info) < 15:
                return None
            atom1 = re.sub("\s", "", hb_info[0:4])
            if hb_info[4] == "-":
                hb_attrbt = "T"
            else:
                hb_attrbt = "P"
            atom2 = re.sub("\s", "", hb_info[5:9])
            hb_dist = float(hb_info[10:])
            hbonds.append((atom1, atom2, hb_dist, hb_attrbt))
        hbonds = tuple(hbonds)
        return hbonds

    def _parse_resinfo(self, res_info, res_num):
        """
        Parse residue information found in base pair and multiplet.

        Arguments:
        o res_info - the list of residue information which is split by ':'
          ## for 3DNA v2.0
          e.g., ....>A  ..70_  [..C]C-----g[.+G]  .106_  B<....
                ...1>A  ...1_  [..C]C-----G[..G]  ..18_  B<...1
                
          ## for 3DNA v1 (obsolete)
          e.g., A .198_ [..A]Ax*--xA[..A] .123_ A
                B .120_ [..U]U
        o res_num - 1/2, single-residue info for multiplet, and
          two-residue info for base pair

        Return values:
        o res_id - tuple (resname, resseq, icode)
        o chain_id - String
        """
        # Residue 1
        model_id1 = re.sub("\.", "", res_info[0][0:4])
        try:
            model_id1 = int(model_id1)
        except ValueError:
            model_id1 = 1
        chain_id1 = re.sub(".*>", "", res_info[0])
        chain_id1 = re.sub("-", " ", chain_id1)
        resseq1 = int(re.sub("\.", "", res_info[1][0:4]))
        icode1 = re.sub("_", " ", res_info[1][4])
        resname1 = re.sub("\.", "", res_info[2][1:4])
        res_id1 = (resname1, resseq1, icode1)
        if res_num == 1:
            return model_id1, chain_id1, res_id1
        # Residue 2
        model_id2 = re.sub("\.", "", res_info[4][2:])
        try:
            model_id2 = int(model_id2)
        except ValueError:
            model_id2 = 1
        chain_id2 = re.sub("<.*", "", res_info[4])
        chain_id2 = re.sub("-", " ", chain_id2)
        resseq2 = int(re.sub("\.", "", res_info[3][0:4]))
        icode2 = re.sub("_", " ", res_info[3][4])
        resname2 = re.sub("\.", "", res_info[2][13:16])
        res_id2 = (resname2, resseq2, icode2)
        return model_id1, chain_id1, res_id1, model_id2, chain_id2, res_id2
        
    def _parse_helix(self, ana_trailer):
        """
        Parse helix related sections in the output file
        (.out) of the command 'analyze'.
        Trailers needed:
        o helix_trailer
        o hbond_trailer
        o bp_trailer
        
        Procedure:
        1. parse helix section to create Element (Helix) and BasepairH
        2. meanwhile, parse hbond and bp_para section
        """
        x3dna_builder = self.x3dna_builder
        structure = self.structure
        helix_trailer, trailer_left = self._get_hlxtrailer(ana_trailer)
        hbond_trailer, trailer_left = self._get_hbtrailer(trailer_left)
        bp_trailer, trailer_left = self._get_bptrailer(trailer_left)
        current_sec_id = None
        helix_counter = 1
        current_helix_seq = 0
        for i in range(0, len(helix_trailer)):
            hlxlist = split(helix_trailer[i])
            helix_sym = hlxlist[4]
            bp_info = split(hlxlist[2], ":")
            if bp_info[2][9] == "-":
                bp_orie = 'anti-parallel'
            else:
                bp_orie = 'parallel'
            # Residues
            model_id1, chain_id1, res_id1, model_id2, chain_id2, res_id2 = self._parse_resinfo(bp_info, 2)
            try:
                residue1 = structure.get_residue(model_id1, chain_id1, res_id1)
                residue1.set_chainno(1)                
            except KeyError:
                raise KeyError, "No <residue %s, %d, %s> found in the structure" % res_id1
            try:
                residue2 = structure.get_residue(model_id2, chain_id2, res_id2)
                residue2.set_chainno(2)
            except KeyError:
                raise KeyError, "No <residue %s, %d, %s> found in the structure" % res_id2
            # Helix id - res_counter1, res_counter2
            hid = (residue1.get_counter(), residue2.get_counter())
            # Hydrogen bonds
            hb_str = hbond_trailer[i][14:]
            hbonds = self._parse_hbinfo(hb_str)
            if hbonds is None:
                continue
            # Base-pair parameters, array
            bplist = split(bp_trailer[i])
            bplist = map(float, bplist[2:])
            bp_para = array(bplist, 'f')
            # Check the helical region
            if helix_sym == "|":
                if current_sec_id != helix_counter:
                    current_sec_id = helix_counter
                    # Initiate helix
                    x3dna_builder.init_helix(current_sec_id, "helix")
                current_helix_seq = current_helix_seq + 1
                # Set helix_seq for residues (used for secondary structure classification)
                residue1.set_helixseq(current_helix_seq)
                residue2.set_helixseq(current_helix_seq)
                # Initiate base pair
                x3dna_builder.init_basepairh(hid, residue1, residue2, bp_para, hbonds, bp_orie, current_helix_seq)
            elif helix_sym == "x":
                current_helix_seq = current_helix_seq + 1
                # Set helix_seq for residues (used for secondary structure classification)
                residue1.set_helixseq(current_helix_seq)
                residue2.set_helixseq(current_helix_seq)
                # Initiate base pair
                x3dna_builder.init_basepairh(hid, residue1, residue2, bp_para, hbonds, bp_orie, current_helix_seq)
                # Increase helix_counter and reset current_helix_seq
                helix_counter = helix_counter + 1
                current_helix_seq = 0
            elif helix_sym == "+":
                current_sec_id = helix_counter
                # Initiate helix
                x3dna_builder.init_helix(current_sec_id, "helix")
                current_helix_seq = 1
                # Set helix_seq for residues (used for secondary structure classification)
                residue1.set_helixseq(current_helix_seq)
                residue2.set_helixseq(current_helix_seq)
                # Initiate base pair
                x3dna_builder.init_basepairh(hid, residue1, residue2, bp_para, hbonds, bp_orie, current_helix_seq)
                # Increase helix_counter and reset current_helix_seq
                helix_counter = helix_counter + 1
                current_helix_seq = 0
        if current_sec_id == None:
            current_sec_id = 0
        return current_sec_id

    def _parse_unpaired(self, last_sec_id):
        "Parse unpaired regions based on helices."
        x3dna_builder = self.x3dna_builder
        current_sec_id = last_sec_id
        current_region = None
        # Get chain list
        chain_list = Selection.unfold_entities(self.structure, "C")
        for chain in chain_list:
            chain.sort()   # Sort the chain by resseq
            for residue in chain.get_iterator():
                if residue.get_ident() == 'Y' or residue.get_ident() == 'R':
                    if residue.get_sec_class() == None:
                        region = "unpaired"
                        if current_region != region:
                            current_region = region
                            current_sec_id = current_sec_id + 1
                            x3dna_builder.init_unpaired(current_sec_id, "unpaired")
                            x3dna_builder.add_residue_to_unpaired(residue)
                        else:
                            x3dna_builder.add_residue_to_unpaired(residue)
                    elif residue.get_sec_class() == "helix":
                        region = "helix"
                        if current_region != region:
                            current_region = region
                    else:
                        raise ParserException("Residue has 'unpaired' secondary label before its definition.")
            current_region = None   # Reset to None

    def _parse_step(self, ana_trailer):
        """
        Parse base-pair step related sections in the output file
        (.out) of the command 'analyze'.
        Trailers needed:
        o step_trailer
        o shelix_trailer
        o phos_trailer
        o groove_trailer
        """
        x3dna_builder = self.x3dna_builder
        step_trailer, trailer_left = self._get_steptrailer(ana_trailer)
        shelix_trailer, trailer_left = self._get_shlxtrailer(trailer_left)
        phos_trailer, trailer_left = self._get_phostrailer(trailer_left)
        groove_trailer, trailer_left = self._get_gtrailer(trailer_left)
        counter_list = []
        step_list = []
        shelix_list = []
        phos_list = []
        form_list = []
        groove_list = []
        for i in range(0, len(step_trailer)):
            slist = split(step_trailer[i])
            # Counter, int
            counter_list.append(int(slist[0]))
            # Step parameters, step_para
            step_para = []
            for value in slist[2:]:
                if value == "---" or value == "----":
                    step_para.append(NullValue)
                else:
                    step_para.append(float(value))
            step_para = array(step_para, 'f')
            step_list.append(step_para)
            # Local base-pair helical parameters, shelix_para
            hlist = split(shelix_trailer[i])
            shelix_para = []
            for value in hlist[2:]:
                if value == "---" or value == "----":
                    shelix_para.append(NullValue)
                else:
                    shelix_para.append(float(value))
            shelix_para = array(shelix_para, 'f')
            shelix_list.append(shelix_para)
            # Phosphate translational parameters, p_para and step_form
            plist = split(phos_trailer[i])
            if len(plist) == 8:   # form not defined
                plist.append(" ")
            step_form = plist[8]
            p_para = []
            for value in plist[2:8]:
                if value == "---" or value == "----":
                    p_para.append(NullValue)
                else:
                    p_para.append(float(value))
            p_para = array(p_para, 'f')
            phos_list.append(p_para)
            form_list.append(step_form)
            # Groove parameters, groove_para
            glist = split(groove_trailer[i])
            groove_para = []
            for value in glist[2:]:
                if value == "---" or value == "----":
                    groove_para.append(NullValue)
                else:
                    groove_para.append(float(value))
            groove_para = array(groove_para, 'f')
            groove_list.append(groove_para)
        # Initiate steplist
        if step_trailer:
            x3dna_builder.set_stepwrapper(counter_list, step_list, shelix_list, phos_list, groove_list, form_list)
        
    def _parse_allbp(self, allbp_trailer):
        """
        Parse all base pairs found by 'find_pair'.
        """
        bpunit_trailer = self._get_bpunit_trailer(allbp_trailer)
        for i in range(0, len(bpunit_trailer), 6):
            bp_unit = bpunit_trailer[i:i+6]
            self._parse_bpunit(bp_unit)        

    def _parse_bpunit(self, bp_unit):
        "Parse base-pair unit (based on 3DNA v2.0)."
        x3dna_builder = self.x3dna_builder
        structure = self.structure
        blist = split(bp_unit[0])
        bp_counter = int(blist[0])
        bp_info = split(blist[4], ":")
        # Residues
        model_id1, chain_id1, res_id1, model_id2, chain_id2, res_id2 = self._parse_resinfo(bp_info, 2)
        try:
            residue1 = structure.get_residue(model_id1, chain_id1, res_id1)
        except KeyError:
            raise KeyError, "No <residue %s> found in the structure" % res_id1
        try:
            residue2 = structure.get_residue(model_id2, chain_id2, res_id2)
        except KeyError:
            raise KeyError, "No <residue %s> found in the structure" % res_id2
        # Orientations
        orlist = split(bp_unit[3])
        z_orie, x_orie, gly_orie = orlist
        dirlist = split(bp_unit[4])
        if dirlist[0][1] == "-":
            y_orie = "trans"
        else:
            y_orie = "cis"
        orie_para = (x_orie, y_orie, z_orie, gly_orie)
        # Set the bp_counter and orientation if they for a base pair embedded in helix
        inhelix = x3dna_builder.chk_basepairh(residue1, residue2, bp_counter, orie_para)
        if not inhelix:  # tertiary interactions
            # Base-pair parameters, array
            para_list = split(bp_unit[1])
            para_list = map(float, para_list[1:])
            bp_para = array(para_list, 'f')
            # Hydrogen bonds
            hb_str = bp_unit[2][17:]
            hbonds = self._parse_hbinfo(hb_str)
            if hbonds is not None:
                self.x3dna_builder.init_basepairu(bp_counter, residue1, residue2, bp_para, hbonds, orie_para)
            
    def _parse_multiple(self, allbp_trailer):
        "Parse multiplet section if any."
        multiper_trailer, multipar_trailer = self._get_multi_trailers(allbp_trailer)
        # Perfect multiplet
        for entry in multiper_trailer:
            self._parse_multi_entry("perfect", entry)
        # Partial multiplet
        for entry in multipar_trailer:
            self._parse_multi_entry("partial", entry)
                
    def _parse_multi_entry(self, quanlity, entry):
        "Parse multi entry."
        structure = self.structure
        res_list = []
        multi_id = []
        mlist = split(entry)
        multi_counter = int(re.sub(":", "", mlist[0]))
        res_num = int(re.sub("#", "", mlist[1]))
        multi_info = re.sub("^\s*\d+:\s#\d+\s+", "", entry)
        resinfo_list = split(multi_info, " + ")
        for i in range(0, res_num):
            res_info = re.sub("^\[\d+\]", "", resinfo_list[i])
            res_info = split(res_info, ":")
            # Get residue
            model_id, chain_id, res_id = self._parse_resinfo(res_info, 1)
            try:
                residue = structure.get_residue(model_id, chain_id, res_id)
            except KeyError:
                raise KeyError, "No <residue %s> found in the structure" % res_id
            res_list.append(residue)
            multi_id.append(residue.get_counter())
        multi_id = tuple(multi_id)   # tuple of residue counters
        self.x3dna_builder.init_multiplet(multi_id, multi_counter, quanlity, res_num, res_list)

    def _parse1(self, allbp_trailer):
        "Parse one output file of the 3DNA program if no helix found."
        x3dna_builder = self.x3dna_builder
        last_sec_id = 0
        self._parse_unpaired(last_sec_id)
        x3dna_builder.set_sec_attrbt()
        self._parse_allbp(allbp_trailer)
        self._parse_multiple(allbp_trailer)
        
    def _parse2(self, ana_trailer, allbp_trailer):
        "Parse two output files of the 3DNA program."
        x3dna_builder = self.x3dna_builder
        last_sec_id = self._parse_helix(ana_trailer)
        self._parse_unpaired(last_sec_id)
        # after getting the complete Secondary instance,
        # set some missing attributes of it, e.g., element_number
        x3dna_builder.set_sec_attrbt()
        self._parse_step(ana_trailer)
        self._parse_allbp(allbp_trailer)
        self._parse_multiple(allbp_trailer)


    # Public methods
    def parse_x3dna(self, bpfile, anafile=None):
        """
        Return the X3DNAInfo object.

        Arguments:
        o bpfile - the output file of 'find_pair'
        o anafile - the output file of 'analyze'

        Procedure:
        1. Parse the anafile to create Secondary and StepList
        2. Parse the bpfile to create Tertiary and MultiList
        """
        x3dna_builder = self.x3dna_builder
        # Create a X3DNA instance
        x3dna_builder.init_x3dna(self.id)
        x3dna_builder.init_secondary("secondary")        
        x3dna_builder.init_tertiary("tertiary")
        x3dna_builder.init_stepwrapper("stepwrapper")
        x3dna_builder.init_multiwrapper("multiwrapper")
        # Get bp_trailer
        bpfh = open(bpfile, 'r')
        allbp_trailer = bpfh.readlines()
        bpfh.close()
        # Check existence of the .out file
        if anafile == None:
            self._parse1(allbp_trailer)
        else:
            anafh = open(anafile, 'r')
            ana_trailer = anafh.readlines()
            anafh.close()
            self._parse2(ana_trailer, allbp_trailer)        

    def parse_rnaview(self, viewfile):
        """
        Parse RNAView file to get LW-base-pairing patterns.
        Arguments:
        o viewfile - the output file of RNAView
        Procedure:
        1. get two residues
        2. find the base pair
        3. assign LW_pattern to the base pair
        4. assign edges to the base pair
        """
        viewfh = open(viewfile)
        allview_trailer = viewfh.readlines()
        viewfh.close()
        for i in range(0, len(allview_trailer)):
            if re.search("^BEGIN_base-pair", allview_trailer[i]):
                begin = i + 1
            if re.search("^END_base-pair", allview_trailer[i]):
                end = i
        bp_trailer = allview_trailer[begin:end]  # base pair information
        for line in bp_trailer:
            if re.search("stacked", line):
                continue
            line_list = line.split()
            # LW pattern
            if re.search("/", line_list[6]) and (line_list[7] == "cis" or line_list[7] == "tran"):
                edge1 = re.sub("\+|-", "W", upper(line_list[6][0]))
                edge2 = re.sub("\+|-", "W", upper(line_list[6][2]))
                orientation = re.sub("tran", "trans", line_list[7])
            else:
                continue            
            # Residue 1
            if line_list[1] == ":":
                chain_id1 = " "
            else:
                chain_id1 = re.sub(":", "", line_list[1])
            res_seq1 = int(line_list[2])
            refname1 = upper(line_list[3][0])
            residue1 = self.structure.find_residue(1, chain_id1, refname1, res_seq1)
            # Residue 2
            if line_list[5] == ":":
                chain_id2 = " "
            else:
                chain_id2 = re.sub(":", "", line_list[5])
            res_seq2 = int(line_list[4])
            refname2 = upper(line_list[3][2])
            residue2 = self.structure.find_residue(1, chain_id2, refname2, res_seq2)
            # Set the LW pattern for the base pair if any
            if (residue1 is not None) and (residue2 is not None):
                self.x3dna_builder.set_LWpat(residue1, residue2, edge1, edge2, orientation)

    def get_x3dna(self):
        "Return the X3DNA instance."
        return self.x3dna_builder.get_x3dna()
