# Created: 10/28/2004
# functions used for parsing 3DNA files

__doc__ = "Functions used for parsing 3DNA files."

import re
import os
import sys
import time
from types import IntType, LongType, FloatType, NoneType, StringType
from string import strip, upper, rstrip, join, lower
from math import degrees, sqrt
from numpy import dot, cross, arccos, float32, float64

from ParserException import ParserException

# Used to represent null number of torsion angle or other parameters
NullValue = 500.0
# Bond upper limit
BondUpperLimit = 2.5

def check_argv(idx, arg):
    """
    Check availability of the argument. Print usage if not available.
    Arguments:
    o idx - index in sys.argv
    o arg - option, e.g., -h
    """
    try:
        sys.argv[idx]
    except ValueError:
        sys.stderr.write("Argument expected for option %s\n" % arg)
        sys.stderr.write(usage)
        sys.exit()
    else:
        return

def check_files(ndb_id, PDIR=None, XDIR=None, VDIR=None):
    """
    Check the ndb path and the existence of the outfile.
    Arguments:
    o ndb_id - NDB id
    Return values:
    o pdbfile - PDB file, in NDB_DATA or current directory
    o bpfile - output file of 'find_pair'
    o outfile - output file of 'analyze'
    o viewfile - output file of RNAview
    """
    pdbfile = ndb_id + ".pdb"
    bpfile = ndb_id + ".bp"
    outfile = ndb_id + ".out"
    viewfile = ndb_id + ".out"
    if PDIR and XDIR and VDIR:  # user defined directories
        pdbfile = PDIR + pdbfile
        bpfile = XDIR + bpfile
        outfile = XDIR + outfile
        viewfile = VDIR + viewfile
    elif os.environ['NDB_DATA']:  # default directory
        NDBDIR = os.environ['NDB_DATA']
        pdbfile = NDBDIR + "/pdbfiles/" + pdbfile
        bpfile = NDBDIR + "/3dna/" + bpfile
        outfile = NDBDIR + "/3dna/" + outfile
        viewfile = NDBDIR + "/rnaview/" + viewfile
    # check the file existence
    if (not os.path.exists(pdbfile)) or (not os.path.exists(bpfile)):
        raise ParserException("Error: no .pdb and .bp files found")
    if not os.path.exists(outfile):
        outfile = ""
    if not os.path.exists(viewfile):
        viewfile = ""
    return pdbfile, bpfile, outfile, viewfile

def get_SPATH():
    "Get the path for X3DNA parser"
    if os.environ["X3DNA_PARSER"]:
        SPATH = os.environ["X3DNA_PARSER"]
        return SPATH
    else:
        raise ParserException("Env variable <X3DNA_PARSER> is not set")

def get_XPATH():
    "Get the path for X3DNA package"
    if os.environ["X3DNAPATH"]:
        XPATH = os.environ["X3DNAPATH"]
        return XPATH
    else:
        raise ParserException("Env variable <X3DNAPATH> is not set")

def contstr(attrbt, precision=None):
    """
    Convert other type to string.
    The main purpose of the function is to check None and NullValue.
    Arguments:
    o attrbt - attribute(s) of objects
    o precision - string of print format, default is '%.1f'
    type:
      o float
      o int (including LongType)
      o string
      o None
      o list
      o tuple
      o array
    Return values:
    o string of the attribute(s)
    """
    floattypes = (FloatType, float32, float64)
    if type(attrbt) is NoneType:
        return '\N'
    elif type(attrbt) is StringType:
        return attrbt
    elif type(attrbt) is IntType or type(attrbt) is LongType:
        return str(attrbt)
    elif type(attrbt) in floattypes:
        if precision == None:
            return "%.1f" % attrbt
        else:
            return precision % attrbt
    else: # list or other similar types
        str_attrbt = []
        for item in attrbt:
            if type(item) in floattypes:
                if item == NullValue:
                    str_attrbt.append('\N')
                else:
                    if precision == None:
                        str_attrbt.append("%.1f" % item)
                    else:
                        str_attrbt.append(precision % item)
            elif type(item) is NoneType:
                str_attrbt.append('\N')
            else:
                str_attrbt.append(str(item))
        return join(str_attrbt, "\t")

def get_basedict():
    """
    Read the <baselist.dat> to get records of three- and one-letter
    base names.
    <baselist.dat> is stored in the same directory
    as a part of the package
    Return:
    o basedict: {original base name: standard base name}
      original base name: upper case, no space, eg. 'ADE'.
      classified name: one letter (upper or lower), eg. 'A' or 'a'.
    """
    SPATH = get_SPATH()
    BPATH = SPATH + "/Data/baselist.dat"
    basefh = open(BPATH, 'r')
    linelist = basefh.readlines()
    basefh.close()
    basedict = {}
    for i in range(0, len(linelist)):
        line = linelist[i]
        line = rstrip(line)   # delete whitespace at the right end
        if re.search("\A#", line) or re.search("\A\s+$", line) or line == "":
            continue
        bname = line[0:3]
        bname = strip(bname)
        bname = upper(bname)
        refname = upper(line[8])
        basedict[bname] = refname
    return basedict

def update_basedict(newbase):
    """
    Update the 'baselist.dat' file.
    o newbase - dict, {modified: (standard name, ndb_id)}
    o Return values: void
    """
    SPATH = get_SPATH()
    BPATH1 = SPATH + "/Data/baselist.dat"
    XPATH = get_XPATH()
    BPATH2 = XPATH + "/config/baselist.dat"
    basefh1 = open(BPATH1, "a")
    basefh2 = open(BPATH2, "a")
    basefh1.write("# Added: %s\n" % time.asctime())
    basefh2.write("# Added: %s\n" % time.asctime())
    for resname in newbase.keys():
        basefh1.write("%3s     %1s      # %s\n" % (resname, lower(newbase[resname][0]), newbase[resname][1]))
        basefh2.write("%3s     %1s      # %s\n" % (resname, lower(newbase[resname][0]), newbase[resname][1]))
    basefh1.close()
    basefh2.close()
    
def switch_hbonds(hbonds):
    """
    Switch atoms in hydrogen bonds and sort.
    Arguments:
    o hbonds - tuple of tuple, (atom1, atom2, ...)
    Return values:
    o reversed hbonds
    """
    new_hb = []
    for hb in hbonds:
        new_hb.append((hb[1], hb[0], hb[2]))
    new_hb.sort()                                 # to order every patterny
    new_hb = tuple(new_hb)
    return new_hb

def get_hbdict(hb_list):
    """
    Get dict of hbonds for each base-pairing pattern.
    Arguments:
    o hblist - hydrogen-bonding records for patterns, tuple of tuple
      (pat_id, atom1, atom2, hb_attrbt)
    Return values:
    o hbond_dict - dict
      pat_name: hbonds
    """
    hbond_dict = {}
    for entry in hb_list:
        pat_name = entry[0]
        if hbond_dict.has_key(pat_name):
            hbond_dict[pat_name].append(entry[1:])
        else:
            hbond_dict[pat_name] = [entry[1:]]
        hbond_dict[pat_name].sort()
    return hbond_dict


# Functions used in structural biology
def virtual_torsion(a, b, c, d):
    """
    Calculate torsion angle (in degree) formed by four points: A-B-C-D.
    Don't check the linkage between atoms.
    The angle is located within [-180, 180].
    1. v_ABC = v_AB x v_BC (cross)
       v_BCD = v_BC x v_CD
    2. theta = arccos(v_ABC * v_BCD) (dot)
    3. determine the sign:
       if v_BC * (v_ABC x v_BCD) > 0, theta > 0
       if v_BC * (v_ABC x v_BCD) < 0, theta < 0
    """
    v_AB = b - a
    v_BC = c - b
    v_CD = d - c
    v_ABC = cross(v_AB, v_BC)
    v_BCD = cross(v_BC, v_CD)
    # Check the cosine range [-1, 1]
    if dot(v_ABC, v_BCD)/(norm(v_ABC) * norm(v_BCD)) < -1.0 or dot(v_ABC, v_BCD)/(norm(v_ABC) * norm(v_BCD)) > 1.0:
        return NullValue
    try:
	theta = arccos(dot(v_ABC, v_BCD)/(norm(v_ABC) * norm(v_BCD)))
    except ZeroDivisionError:
        sys.stderr.write("ZeroDivisionError: vector norm is zero\n")
        return NullValue
    # Determine the sign
    if dot(v_BC, cross(v_ABC, v_BCD)) < 0.000001:
        theta = -theta
    # Convert from radian to degree
    theta = degrees(theta)
    return theta    

def torsion(a, b, c, d):
    """
    Calculate torsion angle (in degree) formed by four points: A-B-C-D.
    Check the linkage between atoms.
    The angle is located within [-180, 180].
    1. v_ABC = v_AB x v_BC (cross)
       v_BCD = v_BC x v_CD
    2. theta = arccos(v_ABC * v_BCD) (dot)
    3. determine the sign:
       if v_BC * (v_ABC x v_BCD) > 0, theta > 0
       if v_BC * (v_ABC x v_BCD) < 0, theta < 0
    Special case: 
    """
    v_AB = b - a
    v_BC = c - b
    v_CD = d - c
    # Check the linkage between atoms
    if norm(v_AB) > BondUpperLimit or norm(v_BC) > BondUpperLimit or norm(v_CD) > BondUpperLimit:
        sys.stderr.write("Error: atoms not linked in a torsion angle\n")
        return NullValue
    v_ABC = cross(v_AB, v_BC)
    v_BCD = cross(v_BC, v_CD)
    # Check the cosine range [-1, 1]
    if dot(v_ABC, v_BCD)/(norm(v_ABC) * norm(v_BCD)) < -1.0 or dot(v_ABC, v_BCD)/(norm(v_ABC) * norm(v_BCD)) > 1.0:
        return NullValue    
    try:
        theta = arccos(dot(v_ABC, v_BCD)/(norm(v_ABC) * norm(v_BCD)))
    except ZeroDivisionError:
        sys.stderr.write("ZeroDivisionError: vector norm is zero\n")
        return NullValue
    # Determine the sign
    if dot(v_BC, cross(v_ABC, v_BCD)) < 0.000001:
        theta = -theta
    # Convert from radian to degree
    theta = degrees(theta)
    return theta

def valence(a, b, c):
    """
    Calculate valence angle between two bonds (vectors) A-B-C.
    The angle is located within [0, 180].
    theta = arccos(v_BC * v_BA)
    """
    v_BA = a - b
    v_BC = c - b
    if dot(v_BA, v_BC)/(norm(v_BA) * norm(v_BC)) < -1.0 or dot(v_BA, v_BC)/(norm(v_BA) * norm(v_BC)) > 1.0:
        return NullValue
    try:
        theta = arccos(dot(v_BA, v_BC)/(norm(v_BA) * norm(v_BC)))
    except ZeroDivisionError:
        sys.stderr.write("ZeroDivisionError: vector norm is zero\n")
        return NullValue
    # Convert from radian to degree
    theta = degrees(theta)
    return theta

def norm(v):
    """
    Get the norm (length) of the vector.
    """
    return sqrt(dot(v, v))


